"""
    This file interfaces with the Lumerical API for modesolving
"""

from emepy.eme import EME
from emepy.fd import ModeSolver, ModeSolver1D
import importlib

if importlib.util.find_spec("lumapi") is not None:
    import lumapi as lm
import numpy as np
import os

from emepy.mode import Mode, Mode1D
from emepy import tools


class LumEME(EME):
    """
    This class is a wrapper for EME, it performs the same operations but uses Lumerical MODE to solve for the modes at the interfaces
    """

    def __init__(self, layers=[], num_periods=1):
        super().__init__(layers, num_periods)

        # open file
        if os.path.isfile("api.lms"):
            os.remove("api.lms")
        self.mode = lm.MODE(hide=True)
        self.mode.save("api.lms")

    def close(self):
        if os.path.isfile("api.lms"):
            os.remove("api.lms")

    def __del__(self):
        if os.path.isfile("api.lms"):
            os.remove("api.lms")

        if os.path.isfile("api_p0.log"):
            os.remove("api_p0.log")


class MSLumerical(ModeSolver):
    """Outdated Lumerical Modesolver. Uses the lumapi Lumerical API. See Modesolver. Parameterizes the cross section as a rectangular waveguide."""

    def __init__(
        self,
        wl=1.55e-6,
        width=0.5e-6,
        thickness=0.22e-6,
        num_modes=1,
        cladding_width=5e-6,
        cladding_thickness=5e-6,
        core_index=None,
        cladding_index=None,
        mesh=300,
        mode=None,
        eme_modes=False,
        polygons=[],
        PML=False,
        **kwargs
    ):
        """MSLumerical class constructor

        Parameters
        ----------
        wl : number
            wavelength of the eigenmodes
        width : number
            width of the core in the cross section
        thickness : number
            thickness of the core in the cross section
        num_modes : int
            number of modes to solve for (default:1)
        cladding_width : number
            width of the cladding in the cross section (default:5e-6)
        cladding_thickness : number
            thickness of the cladding in the cross section (default:5e-6)
        core_index : number
            refractive index of the core (default:Si)
        cladding_index : number
            refractive index of the cladding (default:SiO2)
        mesh : int
            number of mesh points in each direction (xy)
        mode : lumapi.MODE
            MODE object that contains the file information
        eme_modes : boolean
            if true, will utilize the lumerical eme wrapped fde solver which is not normalized to one. Produces slightly different results purely due to roundoff error during normalization.
        PML : boolean
            if true, will enable PML boundary conditions, note: this will increase the mesh and grid space
        """

        # Create global params
        self.wl = wl
        self.width = width
        self.thickness = thickness
        self.num_modes = num_modes
        self.cladding_width = cladding_width
        self.cladding_thickness = cladding_thickness
        self.mesh = mesh
        self.mode = mode
        self.eme_modes = eme_modes
        self.PML = PML
        self.polygons = polygons

        # Check indexes
        self.core_index = tools.Si(wl * 1e6) if core_index is None else core_index
        self.cladding_index = tools.SiO2(wl * 1e6) if cladding_index is None else cladding_index
        self.close_after = False

        # Setup PML
        if self.PML:
            self.setup_PML()
        else:
            self.setup_no_PML()

        # Setup file
        if self.mode is None:
            self.setup_file()

    def setup_PML(self):
        self.num_pml_layers = int(self.mesh / 8.0)
        self.after_x = np.linspace(
            -0.5 * self.cladding_width * (1 + 1 / 4),
            0.5 * self.cladding_width * (1 + 1 / 4),
            self.mesh + 2 * self.num_pml_layers - 1,
        )
        self.after_y = np.linspace(
            -0.5 * self.cladding_width * (1 + 1 / 4),
            0.5 * self.cladding_width * (1 + 1 / 4),
            self.mesh + 2 * self.num_pml_layers - 1,
        )
        self.x = np.linspace(-0.5 * self.cladding_width, 0.5 * self.cladding_width, self.mesh)
        self.y = np.linspace(-0.5 * self.cladding_width, 0.5 * self.cladding_width, self.mesh)

    def setup_file(self):
        self.close_after = True

        # open file
        if os.path.isfile("api.lms"):
            os.remove("api.lms")

        self.mode = lm.MODE(hide=True)
        self.mode.save("api.lms")

    def setup_no_PML(self):
        self.after_x = np.linspace(-0.5 * self.cladding_width, 0.5 * self.cladding_width, self.mesh)
        self.after_y = np.linspace(-0.5 * self.cladding_width, 0.5 * self.cladding_width, self.mesh)
        self.x = self.after_x
        self.y = self.after_y

    def setup_cladding(self):
        cladding = self.mode.addrect()
        cladding.name = "cladding"
        cladding.x = 0
        cladding.x_span = 10.0e-6
        cladding.y = 0
        cladding.y_span = self.cladding_width * 2.0
        cladding.z = 0
        cladding.z_span = self.cladding_thickness * 2.0
        cladding.index = self.cladding_index
        return cladding

    def setup_waveguide_core(self):
        core = self.mode.addrect()
        core.name = "core"
        core.x = 0
        core.x_span = 10.0e-6
        core.y = 0
        core.y_span = self.width
        core.z = 0
        core.z_span = self.thickness
        core.index = self.core_index
        return core

    def setup_polygons_core(self):
        for i, p in enumerate(self.polygons):
            core = self.mode.addpoly()
            self.mode.set("name", str("poly{}".format(i)))
            self.mode.set("vertices", p[0])
            self.mode.set("z span", self.thickness)
            self.mode.set("first axis", "y")
            self.mode.set("rotation 1", 90)
            self.mode.set("y", p[1])
            self.mode.set("z", p[2])
            self.mode.set("index", self.core_index)
        return core

    def setup_eme(self):
        eme = self.mode.addeme()
        self.mode.set("wavelength", self.wl)
        self.mode.set("mesh cells y", self.mesh)
        self.mode.set("mesh cells z", self.mesh)
        self.mode.set("x min", -1e-6)
        self.mode.set("y", 0)
        self.mode.set("y span", self.cladding_width)
        self.mode.set("z", 0)
        self.mode.set("z span", self.cladding_thickness)
        self.mode.set("allow custom eigensolver settings", 1)
        self.mode.set("cells", 1)
        self.mode.set("group spans", 2e-6)
        self.mode.set("modes", self.num_modes)
        if self.PML:
            self.mode.set("y min bc", "PML")
            self.mode.set("y max bc", "PML")
            self.mode.set("z min bc", "PML")
            self.mode.set("z max bc", "PML")
            self.mode.set("pml layers", self.num_pml_layers)
        return eme

    def get_eme_results(self):
        results = self.mode.getresult("EME::Cells::cell_1", "mode fields")
        neffs = []
        gridx, gridy = [results["y"], results["z"]]
        gridx, gridy = [gridx.reshape(gridx.shape[0]), gridy.reshape(gridy.shape[0])]
        mesh = len(gridx)
        fields = []
        for mode_num in range(1, self.num_modes + 1):
            neffs.append(self.mode.getresult("EME::Cells::cell_1", "neff")["neff"].flatten()[mode_num - 1])
            E = results["E" + str(mode_num)].reshape(mesh, mesh, 3)
            H = results["H" + str(mode_num)].reshape(mesh, mesh, 3)
            Ex, Ey, Ez, Hx, Hy, Hz = [E[:, :, 1], E[:, :, 2], E[:, :, 0], H[:, :, 1], H[:, :, 2], H[:, :, 0]]
            fields.append([Hx, Hy, Hz, Ex, Ey, Ez])
        n = results["index"].reshape(mesh, mesh, 3)[:, :, 0]
        return fields, neffs, gridx, gridy, mesh, n

    def setup_fde(self):
        fde = self.mode.addfde()
        fde.y = 0
        fde.y_span = self.cladding_width  # clad_width / 2
        fde.solver_type = "2D X normal"
        fde.x = 0
        fde.z = 0
        fde.z_span = self.cladding_thickness  # clad_thickness / 2
        fde.mesh_cells_y = self.mesh - 1
        fde.mesh_cells_z = self.mesh - 1
        self.mode.set("number of trial modes", self.num_modes)
        self.mode.set("wavelength", self.wl)
        return fde

    def get_fde_results(self):
        self.mode.findmodes()
        gridx, gridy = [self.mode.getresult("FDE::data::mode1", "y"), self.mode.getresult("FDE::data::mode1", "z")]
        gridx, gridy = [gridx.reshape(gridx.shape[0]), gridy.reshape(gridy.shape[0])]
        fields, neffs = [[], []]
        m = self.mesh
        for mode_num in range(1, self.num_modes + 1):
            mode_field = []
            neffs.append(self.mode.getdata("FDE::data::mode" + str(mode_num), "neff")[0][0])
            mode_field.append(self.mode.getdata("FDE::data::mode{}".format(mode_num), "Hy").reshape(m, m))
            mode_field.append(self.mode.getdata("FDE::data::mode{}".format(mode_num), "Hz").reshape(m, m))
            mode_field.append(self.mode.getdata("FDE::data::mode{}".format(mode_num), "Hx").reshape(m, m))
            mode_field.append(self.mode.getdata("FDE::data::mode{}".format(mode_num), "Ey").reshape(m, m))
            mode_field.append(self.mode.getdata("FDE::data::mode{}".format(mode_num), "Ez").reshape(m, m))
            mode_field.append(self.mode.getdata("FDE::data::mode{}".format(mode_num), "Ex").reshape(m, m))
            fields.append(mode_field)
        n = self.mode.getdata("FDE::data::material", "index_x").reshape(m, m)
        return fields, neffs, gridx, gridy, n

    def solve(self):
        """Solves for the eigenmodes"""

        # Restart everything in case this isn't the first go
        self.mode.switchtolayout()
        self.mode.deleteall()

        # Create cladding
        self.cladding = self.setup_cladding()

        # add core x=prop y=width z=thickness
        self.core = self.setup_waveguide_core() if not len(self.polygons) else self.setup_polygons_core()

        # Solve using EME modesolver
        if self.eme_modes or self.PML:

            # set up EME for FDE extraction
            self.eme = self.setup_eme()

            # run
            self.mode.run()
            self.mode.emepropagate()

            # get modes
            fields, neffs, gridx, gridy, mesh, n = self.get_eme_results()
            self.mesh = mesh

        # Solve using FDE modesolver
        else:

            # set up FDE
            self.fde = self.setup_fde()

            # get fde results
            fields, neffs, gridx, gridy, n = self.get_fde_results()

        # Set field, grid, and effective index
        self.x, self.y = [gridx, gridy][:]
        self.neffs, self.fields = [neffs, fields][:]
        self.n = n

        # Check for gui close
        if self.close_after and os.path.isfile("api.lms"):
            os.remove("api.lms")

    def clear(self):
        """Clears the modesolver's eigenmodes to make memory"""

        self.x, self.y, self.neffs, self.fields = [None] * 4

    def get_mode(self, mode_num=0):
        """Get the indexed mode number

        Parameters
        ----------
        mode_num : int
            index of the mode of choice

        Returns
        -------
        Mode
            the eigenmode of index mode_num
        """

        field = self.fields[mode_num]
        Hx, Hy, Hz, Ex, Ey, Ez = field[0:6]
        neff = self.neffs[mode_num]
        mode = Mode(x=self.x, y=self.y, wl=self.wl, neff=neff, Hx=Hx, Hy=Hy, Hz=Hz, Ex=Ex, Ey=Ey, Ez=Ez, n=self.n)

        return mode


class MSLumerical1D(MSLumerical, ModeSolver1D):
    """
    Exact same as MSLumerical except solves on a 1D grid.
    Create geometries in 2D as before, but only the middle y value will be used.
    This declaration will only overload the methods that require changing code for
    """

    def setup_eme(self):
        eme = self.mode.addeme()
        self.mode.set("solver type", "2D XY plane: X prop")
        self.mode.set("wavelength", self.wl)
        self.mode.set("mesh cells y", self.mesh)
        self.mode.set("x min", -1e-6)
        self.mode.set("y", 0)
        self.mode.set("y span", self.cladding_width)
        self.mode.set("z", 0)
        self.mode.set("allow custom eigensolver settings", 1)
        self.mode.set("cells", 1)
        self.mode.set("group spans", 2e-6)
        self.mode.set("modes", self.num_modes)
        if self.PML:
            self.mode.set("y min bc", "PML")
            self.mode.set("y max bc", "PML")
            self.mode.set("pml layers", self.num_pml_layers)
        return eme

    def get_eme_results(self):
        results = self.mode.getresult("EME::Cells::cell_1", "mode fields")
        neffs = []
        gridx = results["y"]
        gridx, gridy = [gridx.reshape(gridx.shape[0]), np.zeros(1)]
        mesh = len(gridx)
        fields = []
        for mode_num in range(1, self.num_modes + 1):
            neffs.append(self.mode.getresult("EME::Cells::cell_1", "neff")["neff"].flatten()[mode_num - 1])
            E = results["E" + str(mode_num)].reshape(mesh, 3)
            H = results["H" + str(mode_num)].reshape(mesh, 3)
            Ex, Ey, Ez, Hx, Hy, Hz = [E[:, 1], E[:, 2], E[:, 0], H[:, 1], H[:, 2], H[:, 0]]
            fields.append([Hx, Hy, Hz, Ex, Ey, Ez])
        n = results["index"].reshape(mesh, 3)[:, 0]
        return fields, neffs, gridx, gridy, mesh, n

    def setup_fde(self):
        fde = self.mode.addfde()
        fde.y = 0
        fde.y_span = self.cladding_width  # clad_width / 2
        fde.solver_type = "1D Y:X prop"
        fde.x = 0
        fde.z = 0
        fde.mesh_cells_y = self.mesh - 1
        self.mode.set("number of trial modes", self.num_modes)
        self.mode.set("wavelength", self.wl)
        return fde

    def get_fde_results(self):
        self.mode.findmodes()
        gridx = self.mode.getresult("FDE::data::mode1", "y")
        gridx = gridx.reshape(gridx.shape[0])
        gridy = np.zeros(1)
        fields, neffs = [[], []]
        m = self.mesh
        for mode_num in range(1, self.num_modes + 1):
            mode_field = []
            neffs.append(self.mode.getdata("FDE::data::mode" + str(mode_num), "neff")[0][0])
            mode_field.append(self.mode.getdata("FDE::data::mode{}".format(mode_num), "Hy").reshape(m))
            mode_field.append(self.mode.getdata("FDE::data::mode{}".format(mode_num), "Hz").reshape(m))
            mode_field.append(self.mode.getdata("FDE::data::mode{}".format(mode_num), "Hx").reshape(m))
            mode_field.append(self.mode.getdata("FDE::data::mode{}".format(mode_num), "Ey").reshape(m))
            mode_field.append(self.mode.getdata("FDE::data::mode{}".format(mode_num), "Ez").reshape(m))
            mode_field.append(self.mode.getdata("FDE::data::mode{}".format(mode_num), "Ex").reshape(m))
            fields.append(mode_field)
        n = self.mode.getdata("FDE::data::material", "index_x").reshape(m)
        return fields, neffs, gridx, gridy, n

    def get_mode(self, mode_num=0):
        field = self.fields[mode_num]
        Hx, Hy, Hz, Ex, Ey, Ez = field[0:6]
        neff = self.neffs[mode_num]
        mode = Mode1D(x=self.x, wl=self.wl, neff=neff, Hx=Hx, Hy=Hy, Hz=Hz, Ex=Ex, Ey=Ey, Ez=Ez, n=self.n)
        return mode

    def setup_PML(self):
        self.num_pml_layers = int(self.mesh / 8.0)
        self.after_x = np.linspace(
            -0.5 * self.cladding_width * (1 + 1 / 4),
            0.5 * self.cladding_width * (1 + 1 / 4),
            self.mesh + 2 * self.num_pml_layers - 1,
        )
        self.after_y = np.zeros(1)
        self.x = np.linspace(-0.5 * self.cladding_width, 0.5 * self.cladding_width, self.mesh)
        self.y = np.zeros(1)

    def setup_no_PML(self):
        self.after_x = np.linspace(-0.5 * self.cladding_width, 0.5 * self.cladding_width, self.mesh)
        self.after_y = np.zeros(1)
        self.x = self.after_x
        self.y = self.after_y



import os
import numpy as np
import pickle

import torch
import torch.nn as nn

from sklearn.preprocessing import PolynomialFeatures
import sklearn

from emepy.mode import Mode, EigenMode
from emepy.fd import ModeSolver, EMpy_gpu
from emepy.tools import from_chunks, get_epsfunc, _get_eps, interp
from emepy.materials import Si, SiO2
from scipy.sparse import coo_matrix

from typing import Callable

FIELD_WIDTH = 128
FIELD_SIZE = FIELD_WIDTH ** 2


def getUpConvLayer(i_size: int, o_size: int, kernal: int, channels: int, first: bool = False, last: bool = False):
    """Returns the right size for up convolutional sampling"""

    def out_size(in_size, kernal, stride, padding, output_padding):
        return (in_size - 1) * stride - 2 * padding + kernal + output_padding

    stride = 0
    padding = 0
    output_padding = 0

    if not out_size(i_size, kernal, stride, padding, output_padding) == o_size:
        while out_size(i_size, kernal, stride, padding, output_padding) < o_size:
            stride += 1

    if not out_size(i_size, kernal, stride, padding, output_padding) == o_size:
        while out_size(i_size, kernal, stride, padding, output_padding) > o_size:
            padding += 1

    if not out_size(i_size, kernal, stride, padding, output_padding) == o_size:
        while out_size(i_size, kernal, stride, padding, output_padding) < o_size:
            output_padding += 1

    if first:
        return nn.ConvTranspose2d(
            1, channels, stride=stride, padding=padding, kernel_size=kernal, output_padding=output_padding
        )

    if last:
        return nn.ConvTranspose2d(
            channels, 1, stride=stride, padding=padding, kernel_size=kernal, output_padding=output_padding
        )

    return nn.ConvTranspose2d(
        channels, channels, stride=stride, padding=padding, kernel_size=kernal, output_padding=output_padding
    )


def getDownConvLayer(i_size: int, o_size: int, kernal: int, channels: int, first: bool = False, last: bool = False):
    """Returns the right size for down convolutional sampling"""

    def out_size(in_size, kernal, stride, padding, dilation):
        return (in_size + 2 * padding - dilation * (kernal - 1) - 1) / stride + 1

    stride = 1
    padding = 0
    dilation = 1

    if not out_size(i_size, kernal, stride, padding, dilation) == o_size:
        while out_size(i_size, kernal, stride, padding, dilation) > o_size:
            dilation += 1

    if not out_size(i_size, kernal, stride, padding, dilation) == o_size:
        while out_size(i_size, kernal, stride, padding, dilation) < o_size:
            padding += 1

    if not out_size(i_size, kernal, stride, padding, dilation) == o_size:
        raise Exception("Choose a different kernal size")

    if first:
        return nn.Conv2d(1, channels, stride=stride, padding=padding, kernel_size=kernal, dilation=dilation)

    if last:
        return nn.Conv2d(channels, 1, stride=stride, padding=padding, kernel_size=kernal, dilation=dilation)

    return nn.Conv2d(channels, channels, stride=stride, padding=padding, kernel_size=kernal, dilation=dilation)


class Network(nn.Module):
    """The pytorch inherited class that defines and represents the physical neural network"""

    def __init__(self, code_size: int, channels: int, component: str) -> None:
        """Network constructor

        Parameters
        ----------
        code_size : int
            the number of inputs
        channels : int
            the number of channels
        component : string
            "Hx" or "Hy"
        """
        super().__init__()
        self.channels = channels

        self.linear_up_1 = nn.Linear(code_size, int(FIELD_WIDTH / 20) ** 2)
        self.linear_up_2 = nn.Linear(int(FIELD_WIDTH / 20) ** 2, int(FIELD_WIDTH / 7) ** 2)
        self.linear_up_3 = nn.Linear(int(FIELD_WIDTH / 7) ** 2, int(FIELD_WIDTH / 4) ** 2)
        self.linear_up_4 = nn.Linear(int(FIELD_WIDTH / 4) ** 2, int(FIELD_WIDTH / 3) ** 2)
        self.conv_up_1 = getUpConvLayer(int(FIELD_WIDTH / 3), int(FIELD_WIDTH / 2), 3, channels, first=True)
        self.conv_up_2 = getUpConvLayer(int(FIELD_WIDTH / 2), int(5 * FIELD_WIDTH / 8), 5, channels)
        self.conv_up_3 = getUpConvLayer(int(5 * FIELD_WIDTH / 8), int(6 * FIELD_WIDTH / 8), 7, channels)
        self.conv_up_4 = getUpConvLayer(int(6 * FIELD_WIDTH / 8), int(7 * FIELD_WIDTH / 8), 7, channels)
        self.conv_up_5 = getUpConvLayer(int(7 * FIELD_WIDTH / 8), int(FIELD_WIDTH), 9, channels, last=True)
        self.linear_up_5 = nn.Linear(FIELD_WIDTH ** 2, FIELD_WIDTH ** 2)

        self.relu = nn.ReLU()
        self.tanh = nn.Tanh()
        self.sigmoid = nn.Sigmoid()

        torch.nn.init.xavier_normal_(self.linear_up_1.weight)
        torch.nn.init.xavier_normal_(self.linear_up_2.weight)
        torch.nn.init.xavier_normal_(self.linear_up_3.weight)

        torch.nn.init.xavier_normal_(self.conv_up_1.weight)
        torch.nn.init.xavier_normal_(self.conv_up_2.weight)
        torch.nn.init.xavier_normal_(self.conv_up_3.weight)

        self.component = component

    def forward(self, field) -> tuple:
        """Performs the network propagation

        Parameters
        ----------
        field : array
            The inputs to the network

        Returns
        -------
        tuple (torch array, torch array)
            Returns a tuple of the outputs, inputs
        """

        out = self.tanh(self.linear_up_1(field)).view(-1, 1, int(FIELD_WIDTH / 20) ** 2)
        out = self.tanh(self.linear_up_2(out)).view(-1, 1, int(FIELD_WIDTH / 7) ** 2)
        out = self.tanh(self.linear_up_3(out)).view(-1, 1, int(FIELD_WIDTH / 4) ** 2)
        out = self.tanh(self.linear_up_4(out)).view(-1, 1, int(FIELD_WIDTH / 3), int(FIELD_WIDTH / 3))
        out = self.tanh(self.conv_up_1(out)).view(-1, self.channels, int(FIELD_WIDTH / 2), int(FIELD_WIDTH / 2))
        out = self.tanh(self.conv_up_2(out)).view(-1, self.channels, int(5 * FIELD_WIDTH / 8), int(5 * FIELD_WIDTH / 8))
        out = self.tanh(self.conv_up_3(out)).view(-1, self.channels, int(6 * FIELD_WIDTH / 8), int(6 * FIELD_WIDTH / 8))
        out = self.tanh(self.conv_up_4(out)).view(-1, self.channels, int(7 * FIELD_WIDTH / 8), int(7 * FIELD_WIDTH / 8))
        out = self.tanh(self.conv_up_5(out)).view(-1, 1, FIELD_WIDTH ** 2)

        out = self.linear_up_5(out).view(-1, FIELD_WIDTH, FIELD_WIDTH)

        out = out / 1000.0 if self.component == "Hx" else out / 100.0

        return out, field


class ANN(object):
    """Object that loads the neural network; Users are heavily encouraged to design their own networks and rewrite their own ANN to match their needs"""

    def __init__(self,) -> None:
        """Constructor for Mode Object"""
        self.x = np.linspace(0, 2.5, FIELD_WIDTH)
        self.y = np.linspace(0, 2.5, FIELD_WIDTH)

        self.Hx_model = self.Hx_network()
        self.Hy_model = self.Hy_network()
        self.neff_model = self.neff_regression()

    def neff_regression(self) -> "sklearn.linear_model._base.LinearRegression":
        """Return the opened regression model for the effective index"""

        with open(os.path.dirname(os.path.abspath(__file__)) + "/models/neff_pickle/model.pk", "rb") as f:
            model = pickle.load(f)

        return model

    def Hx_network(self) -> Network:
        """Return the opened network model for the Hx component"""

        from_chunks(os.path.dirname(os.path.abspath(__file__)) + "/models/Hx_chunks/", "hx_temp.pt")
        with open("hx_temp.pt", "rb") as f:
            model = Network(3, 1, "Hx")

            # original saved file with DataParallel
            state_dict = torch.load(f)
            model.load_state_dict(state_dict)

            model.eval()

        os.system("rm hx_temp.pt")
        return model

    def Hy_network(self) -> Network:
        """Return the opened network model for the Hy component"""

        from_chunks(os.path.dirname(os.path.abspath(__file__)) + "/models/Hy_chunks/", "hy_temp.pt")
        with open("hy_temp.pt", "rb") as f:
            model = Network(3, 1, "Hy")

            # original saved file with DataParallel
            state_dict = torch.load(f)
            model.load_state_dict(state_dict)

            model.eval()

        os.system("rm hy_temp.pt")
        return model


class MSNeuralNetwork(ModeSolver):
    """ModeSolver object for the sample neural networks, parameterizes the cross section components. Currently designed only for single mode calculations in Silicon on SiO2"""

    def __init__(self, ann: ANN, wl: float, width: float, thickness: float) -> None:
        """MSNeuralNetwork constructor

        Parameters
        ----------
        ann : ANN
            The ANN object that contains the network and regression models
        wl : number
            The wavelength (most accurate around 1.55 µm)
        width : number
            The width of the cross section (most accurate around 550 nm)
        thickness : number
            The thickness of the cross section (most accurate around 250 nm)
        """

        self.wl = wl
        self.width = width
        self.thickness = thickness
        self.ann = ann
        self.Hx_model = ann.Hx_model
        self.Hy_model = ann.Hy_model
        self.neff_model = ann.neff_model
        self.num_modes = 1
        self.x = ann.x
        self.y = ann.y
        self.after_x = self.x
        self.after_y = self.y
        self.mesh = len(self.x) - 1
        self.PML = False
        self.n = get_epsfunc(self.width, self.thickness, 2.5, 2.5, Si(self.wl), SiO2(self.wl), compute=True)(
            self.x, self.y
        )

    def solve(self) -> None:
        """Solves for the eigenmode using the neural networks"""

        self.mode = None

        Hx, Hy, neff = self.data(self.width, self.thickness, self.wl)
        self.mode = (Hx, Hy, neff)

    def data(self, width: float, thickness: float, wl: float) -> tuple:
        """Propagates the inputs into the neural networks and regression models and returns the outputs

        Parameters
        ----------
        width : number
            The width of the cross section (most accurate around 550 nm)
        thickness : number
            The thickness of the cross section (most accurate around 250 nm)
        wl : number
            The wavelength (most accurate around 1.55 µm)

        Returns
        -------
        tuple (numpy array, numpy array, number)
            Returns Hx, Hy, and neff
        """

        neff = self.neff_regression(width, thickness, wl, self.neff_model)
        Hx = self.Hx_network(width, thickness, wl, self.Hx_model)
        Hy = self.Hy_network(width, thickness, wl, self.Hy_model)

        return Hx, Hy, neff

    def neff_regression(
        self, width: float, thickness: float, wl: float, model: "sklearn.linear_model._base.LinearRegression"
    ) -> float:
        """Calculates the effective index using a regression model

        Parameters
        ----------
        width : number
            The width of the cross section (most accurate around 550 nm)
        thickness : number
            The thickness of the cross section (most accurate around 250 nm)
        wl : number
            The wavelength (most accurate around 1.55 µm)
        model : sklearn regression model
            The model that performs the regression

        Returns
        -------
        number
            Returns the effective index
        """

        poly = PolynomialFeatures(degree=8)
        X = poly.fit_transform([[width, thickness, wl]])
        neff = model.predict(X)

        return neff[0]

    def Hx_network(self, width: float, thickness: float, wl: float, model: Network) -> "np.ndarray":
        """Calculates the Hx component using a network model

        Parameters
        ----------
        width : number
            The width of the cross section (most accurate around 550 nm)
        thickness : number
            The thickness of the cross section (most accurate around 250 nm)
        wl : number
            The wavelength (most accurate around 1.55 µm)
        model : pytorch model
            The model that performs the ann calculation

        Returns
        -------
        numpy array
            Returns the Hx field
        """

        with torch.no_grad():
            parameters = torch.Tensor([[[width, thickness, wl]]])
            output, _ = model(parameters)
            output = output.numpy()
            output = output.reshape(128, 128)

        return output

    def Hy_network(self, width: float, thickness: float, wl: float, model: Network) -> "np.ndarray":
        """Calculates the Hy component using a network model

        Parameters
        ----------
        width : number
            The width of the cross section (most accurate around 550 nm)
        thickness : number
            The thickness of the cross section (most accurate around 250 nm)
        wl : number
            The wavelength (most accurate around 1.55 µm)
        model : pytorch model
            The model that performs the ann calculation

        Returns
        -------
        numpy array
            Returns the Hy field
        """

        with torch.no_grad():
            parameters = torch.Tensor([[[width, thickness, wl]]])
            output, _ = model(parameters)
            output = output.numpy()
            output = output.reshape(128, 128)

        return output

    def clear(self) -> None:
        """Clears the mode in the object"""
        self.mode = None

    def get_mode(self, mode_num: int = 0) -> EigenMode:
        """Returns the solved eigenmode

        Parameters
        ----------
        mode_num : int
            mode index to return mode of

        Returns
        -------
        EigenMode
            the EigenMode corresponding to the provdided mode index
        """
        Hx, Hy, neff = self.mode

        epsfunc_before = get_epsfunc(self.width, self.thickness, 2.5, 2.5, Si(self.wl), SiO2(self.wl), compute=True)
        epsfunc_after = get_epsfunc(self.width, self.thickness, 2.5, 2.5, Si(self.wl), SiO2(self.wl))
        m = Mode(
            self.x,
            self.y,
            self.wl,
            neff,
            Hx + 0j,
            Hy + 0j,
            None,
            None,
            None,
            None,
            np.sqrt(epsfunc_after(self.x, self.y)),
        )
        m.compute_other_fields(epsfunc_before, epsfunc_after)
        m.normalize()

        return m


def compute_other_fields(
    mode,
    epsfunc_1: Callable[["np.ndarray", "np.ndarray"], "np.ndarray"],
    epsfunc_2: Callable[["np.ndarray", "np.ndarray"], "np.ndarray"],
    boundary: str = "0000",
) -> None:
    """Given the Hx and Hy fields, maxwell's curl relations can be used to calculate the remaining field; adapted from the EMpy

    Parameters
    ----------
    epsfunc_1 : function
        epsfunc for computing other fields
    epsfunc_2 : function
        epsfunc for computing final refractive index profile
    boundary : str
        the boundary conditions as defined by electromagneticpython
    """

    (
        mode.Hx,
        mode.Hy,
        mode.Hz,
        mode.Ex,
        mode.Ey,
        mode.Ez,
    ) = compute_other_fields_2D(
        mode.neff, mode.Hx, mode.Hy, mode.wl, mode.x, mode.y, boundary, epsfunc_1
    )
    x_ = (mode.x[1:] + mode.x[:-1]) / 2.0
    y_ = (mode.y[1:] + mode.y[:-1]) / 2.0
    mode.Ex = interp(mode.x, mode.y, x_, y_, mode.Ex, False)
    mode.Ey = interp(mode.x, mode.y, x_, y_, mode.Ey, False)
    mode.Ez = interp(mode.x, mode.y, x_, y_, mode.Ez, False)
    mode.x = mode.x - mode.x[int(len(mode.x) / 2)]
    mode.y = mode.y - mode.y[int(len(mode.y) / 2)]
    mode.H = np.sqrt(
        np.abs(mode.Hx) ** 2 + np.abs(mode.Hy) ** 2 + np.abs(mode.Hz) ** 2
    )
    mode.E = np.sqrt(
        np.abs(mode.Ex) ** 2 + np.abs(mode.Ey) ** 2 + np.abs(mode.Ez) ** 2
    )
    mode.n = np.sqrt(epsfunc_2(mode.x, mode.y))


def compute_other_fields_2D(
    neff: float,
    Hx: "np.ndarray",
    Hy: "np.ndarray",
    wl: float,
    x: "np.ndarray",
    y: "np.ndarray",
    boundary: str,
    epsfunc: Callable[["np.ndarray", "np.ndarray"], "np.ndarray"],
) -> list:
    """Given the Hx and Hy fields, will calculate the rest

    Parameters
    ----------
    neff: float
        the effective index of the EigenMode
    Hx: "np.ndarray"
        the Hx field profile
    Hy: "np.ndarray"
        the Hy field profile
    wl: float
        the wavelength of the simulation
    x: "np.ndarray"
        the x numpy grid array
    y: "np.ndarray"
        the y numpy grid array
    boundary: str
        a string representing the boundary conditions as defined for the electromagneticpython library
    epsfunc: function
        a function of (x,y) that can be called to return the refractive index field profile for the domain

    Returns
    -------
    list
        a list of the field profile np.ndarrays [Hx,Hy,Hz,Ex,Ey,Ez]
    """

    dx = np.diff(x)
    dy = np.diff(y)

    dx = np.r_[dx[0], dx, dx[-1]].reshape(-1, 1)
    dy = np.r_[dy[0], dy, dy[-1]].reshape(1, -1)

    xc = (x[:-1] + x[1:]) / 2
    yc = (y[:-1] + y[1:]) / 2
    epsxx, epsxy, epsyx, epsyy, epszz = _get_eps(xc, yc, epsfunc)

    nx = len(x)
    ny = len(y)

    k = 2 * np.pi / wl

    ones_nx = np.ones((nx, 1))
    ones_ny = np.ones((1, ny))

    n = np.dot(ones_nx, dy[:, 1:]).flatten()
    s = np.dot(ones_nx, dy[:, :-1]).flatten()
    e = np.dot(dx[1:, :], ones_ny).flatten()
    w = np.dot(dx[:-1, :], ones_ny).flatten()

    exx1 = epsxx[:-1, 1:].flatten()
    exx2 = epsxx[:-1, :-1].flatten()
    exx3 = epsxx[1:, :-1].flatten()
    exx4 = epsxx[1:, 1:].flatten()

    eyy1 = epsyy[:-1, 1:].flatten()
    eyy2 = epsyy[:-1, :-1].flatten()
    eyy3 = epsyy[1:, :-1].flatten()
    eyy4 = epsyy[1:, 1:].flatten()

    exy1 = epsxy[:-1, 1:].flatten()
    exy2 = epsxy[:-1, :-1].flatten()
    exy3 = epsxy[1:, :-1].flatten()
    exy4 = epsxy[1:, 1:].flatten()

    eyx1 = epsyx[:-1, 1:].flatten()
    eyx2 = epsyx[:-1, :-1].flatten()
    eyx3 = epsyx[1:, :-1].flatten()
    eyx4 = epsyx[1:, 1:].flatten()

    ezz1 = epszz[:-1, 1:].flatten()
    ezz2 = epszz[:-1, :-1].flatten()
    ezz3 = epszz[1:, :-1].flatten()
    ezz4 = epszz[1:, 1:].flatten()

    b = neff * k

    bzxne = (
        0.5
        * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
        * eyx4
        / ezz4
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy3
        * eyy1
        * w
        * eyy2
        + 0.5
        * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
        * (1 - exx4 / ezz4)
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * exx1
        * s
    ) / b

    bzxse = (
        -0.5
        * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
        * eyx3
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy1
        * w
        * eyy2
        + 0.5
        * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
        * (1 - exx3 / ezz3)
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * n
        * exx1
        * exx4
    ) / b

    bzxnw = (
        -0.5
        * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
        * eyx1
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy2
        * e
        - 0.5
        * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
        * (1 - exx1 / ezz1)
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * exx4
        * s
    ) / b

    bzxsw = (
        0.5
        * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
        * eyx2
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * e
        - 0.5
        * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
        * (1 - exx2 / ezz2)
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx3
        * n
        * exx1
        * exx4
    ) / b

    bzxn = (
        (
            0.5
            * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
            * n
            * ezz1
            * ezz2
            / eyy1
            * (2 * eyy1 / ezz1 / n ** 2 + eyx1 / ezz1 / n / w)
            + 0.5
            * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
            * n
            * ezz4
            * ezz3
            / eyy4
            * (2 * eyy4 / ezz4 / n ** 2 - eyx4 / ezz4 / n / e)
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
            * (
                0.5
                * ezz4
                * ((1 - exx1 / ezz1) / n / w - exy1 / ezz1 * (2.0 / n ** 2 - 2 / n ** 2 * s / (n + s)))
                / exx1
                * ezz1
                * w
                + (ezz4 - ezz1) * s / n / (n + s)
                + 0.5
                * ezz1
                * (-(1 - exx4 / ezz4) / n / e - exy4 / ezz4 * (2.0 / n ** 2 - 2 / n ** 2 * s / (n + s)))
                / exx4
                * ezz4
                * e
            )
            - (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
            * (
                -ezz3 * exy2 / n / (n + s) / exx2 * w
                + (ezz3 - ezz2) * s / n / (n + s)
                - ezz2 * exy3 / n / (n + s) / exx3 * e
            )
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzxs = (
        (
            0.5
            * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
            * s
            * ezz2
            * ezz1
            / eyy2
            * (2 * eyy2 / ezz2 / s ** 2 - eyx2 / ezz2 / s / w)
            + 0.5
            * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
            * s
            * ezz3
            * ezz4
            / eyy3
            * (2 * eyy3 / ezz3 / s ** 2 + eyx3 / ezz3 / s / e)
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
            * (
                -ezz4 * exy1 / s / (n + s) / exx1 * w
                - (ezz4 - ezz1) * n / s / (n + s)
                - ezz1 * exy4 / s / (n + s) / exx4 * e
            )
            - (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
            * (
                0.5
                * ezz3
                * (-(1 - exx2 / ezz2) / s / w - exy2 / ezz2 * (2.0 / s ** 2 - 2 / s ** 2 * n / (n + s)))
                / exx2
                * ezz2
                * w
                - (ezz3 - ezz2) * n / s / (n + s)
                + 0.5
                * ezz2
                * ((1 - exx3 / ezz3) / s / e - exy3 / ezz3 * (2.0 / s ** 2 - 2 / s ** 2 * n / (n + s)))
                / exx3
                * ezz3
                * e
            )
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzxe = (
        (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
        * (
            0.5 * n * ezz4 * ezz3 / eyy4 * (2.0 / e ** 2 - eyx4 / ezz4 / n / e)
            + 0.5 * s * ezz3 * ezz4 / eyy3 * (2.0 / e ** 2 + eyx3 / ezz3 / s / e)
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            -0.5 * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e) * ezz1 * (1 - exx4 / ezz4) / n / exx4 * ezz4
            - 0.5 * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e) * ezz2 * (1 - exx3 / ezz3) / s / exx3 * ezz3
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzxw = (
        (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
        * (
            0.5 * n * ezz1 * ezz2 / eyy1 * (2.0 / w ** 2 + eyx1 / ezz1 / n / w)
            + 0.5 * s * ezz2 * ezz1 / eyy2 * (2.0 / w ** 2 - eyx2 / ezz2 / s / w)
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            0.5 * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e) * ezz4 * (1 - exx1 / ezz1) / n / exx1 * ezz1
            + 0.5 * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e) * ezz3 * (1 - exx2 / ezz2) / s / exx2 * ezz2
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzxp = (
        (
            (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
            * (
                0.5
                * n
                * ezz1
                * ezz2
                / eyy1
                * (-2.0 / w ** 2 - 2 * eyy1 / ezz1 / n ** 2 + k ** 2 * eyy1 - eyx1 / ezz1 / n / w)
                + 0.5
                * s
                * ezz2
                * ezz1
                / eyy2
                * (-2.0 / w ** 2 - 2 * eyy2 / ezz2 / s ** 2 + k ** 2 * eyy2 + eyx2 / ezz2 / s / w)
            )
            + (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
            * (
                0.5
                * n
                * ezz4
                * ezz3
                / eyy4
                * (-2.0 / e ** 2 - 2 * eyy4 / ezz4 / n ** 2 + k ** 2 * eyy4 + eyx4 / ezz4 / n / e)
                + 0.5
                * s
                * ezz3
                * ezz4
                / eyy3
                * (-2.0 / e ** 2 - 2 * eyy3 / ezz3 / s ** 2 + k ** 2 * eyy3 - eyx3 / ezz3 / s / e)
            )
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
            * (
                0.5
                * ezz4
                * (
                    -(k ** 2) * exy1
                    - (1 - exx1 / ezz1) / n / w
                    - exy1 / ezz1 * (-2.0 / n ** 2 - 2 / n ** 2 * (n - s) / s)
                )
                / exx1
                * ezz1
                * w
                + (ezz4 - ezz1) * (n - s) / n / s
                + 0.5
                * ezz1
                * (
                    -(k ** 2) * exy4
                    + (1 - exx4 / ezz4) / n / e
                    - exy4 / ezz4 * (-2.0 / n ** 2 - 2 / n ** 2 * (n - s) / s)
                )
                / exx4
                * ezz4
                * e
            )
            - (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
            * (
                0.5
                * ezz3
                * (
                    -(k ** 2) * exy2
                    + (1 - exx2 / ezz2) / s / w
                    - exy2 / ezz2 * (-2.0 / s ** 2 + 2 / s ** 2 * (n - s) / n)
                )
                / exx2
                * ezz2
                * w
                + (ezz3 - ezz2) * (n - s) / n / s
                + 0.5
                * ezz2
                * (
                    -(k ** 2) * exy3
                    - (1 - exx3 / ezz3) / s / e
                    - exy3 / ezz3 * (-2.0 / s ** 2 + 2 / s ** 2 * (n - s) / n)
                )
                / exx3
                * ezz3
                * e
            )
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzyne = (
        0.5
        * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
        * (1 - eyy4 / ezz4)
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy3
        * eyy1
        * w
        * eyy2
        + 0.5
        * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
        * exy4
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * exx1
        * s
    ) / b

    bzyse = (
        -0.5
        * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
        * (1 - eyy3 / ezz3)
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy1
        * w
        * eyy2
        + 0.5
        * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
        * exy3
        / ezz3
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * n
        * exx1
        * exx4
    ) / b

    bzynw = (
        -0.5
        * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
        * (1 - eyy1 / ezz1)
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy2
        * e
        - 0.5
        * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
        * exy1
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * exx4
        * s
    ) / b

    bzysw = (
        0.5
        * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
        * (1 - eyy2 / ezz2)
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * e
        - 0.5
        * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
        * exy2
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx3
        * n
        * exx1
        * exx4
    ) / b

    bzyn = (
        (
            0.5 * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3) * ezz1 * ezz2 / eyy1 * (1 - eyy1 / ezz1) / w
            - 0.5 * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2) * ezz4 * ezz3 / eyy4 * (1 - eyy4 / ezz4) / e
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
        * (
            0.5 * ezz4 * (2.0 / n ** 2 + exy1 / ezz1 / n / w) / exx1 * ezz1 * w
            + 0.5 * ezz1 * (2.0 / n ** 2 - exy4 / ezz4 / n / e) / exx4 * ezz4 * e
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzys = (
        (
            -0.5 * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3) * ezz2 * ezz1 / eyy2 * (1 - eyy2 / ezz2) / w
            + 0.5 * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2) * ezz3 * ezz4 / eyy3 * (1 - eyy3 / ezz3) / e
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        - (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
        * (
            0.5 * ezz3 * (2.0 / s ** 2 - exy2 / ezz2 / s / w) / exx2 * ezz2 * w
            + 0.5 * ezz2 * (2.0 / s ** 2 + exy3 / ezz3 / s / e) / exx3 * ezz3 * e
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzye = (
        (
            (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
            * (
                -n * ezz2 / eyy1 * eyx1 / e / (e + w)
                + (ezz1 - ezz2) * w / e / (e + w)
                - s * ezz1 / eyy2 * eyx2 / e / (e + w)
            )
            + (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
            * (
                0.5
                * n
                * ezz4
                * ezz3
                / eyy4
                * (-(1 - eyy4 / ezz4) / n / e - eyx4 / ezz4 * (2.0 / e ** 2 - 2 / e ** 2 * w / (e + w)))
                + 0.5
                * s
                * ezz3
                * ezz4
                / eyy3
                * ((1 - eyy3 / ezz3) / s / e - eyx3 / ezz3 * (2.0 / e ** 2 - 2 / e ** 2 * w / (e + w)))
                + (ezz4 - ezz3) * w / e / (e + w)
            )
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            0.5
            * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
            * ezz1
            * (2 * exx4 / ezz4 / e ** 2 - exy4 / ezz4 / n / e)
            / exx4
            * ezz4
            * e
            - 0.5
            * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
            * ezz2
            * (2 * exx3 / ezz3 / e ** 2 + exy3 / ezz3 / s / e)
            / exx3
            * ezz3
            * e
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzyw = (
        (
            (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
            * (
                0.5
                * n
                * ezz1
                * ezz2
                / eyy1
                * ((1 - eyy1 / ezz1) / n / w - eyx1 / ezz1 * (2.0 / w ** 2 - 2 / w ** 2 * e / (e + w)))
                - (ezz1 - ezz2) * e / w / (e + w)
                + 0.5
                * s
                * ezz2
                * ezz1
                / eyy2
                * (-(1 - eyy2 / ezz2) / s / w - eyx2 / ezz2 * (2.0 / w ** 2 - 2 / w ** 2 * e / (e + w)))
            )
            + (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
            * (
                -n * ezz3 / eyy4 * eyx4 / w / (e + w)
                - s * ezz4 / eyy3 * eyx3 / w / (e + w)
                - (ezz4 - ezz3) * e / w / (e + w)
            )
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            0.5
            * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
            * ezz4
            * (2 * exx1 / ezz1 / w ** 2 + exy1 / ezz1 / n / w)
            / exx1
            * ezz1
            * w
            - 0.5
            * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
            * ezz3
            * (2 * exx2 / ezz2 / w ** 2 - exy2 / ezz2 / s / w)
            / exx2
            * ezz2
            * w
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzyp = (
        (
            (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
            * (
                0.5
                * n
                * ezz1
                * ezz2
                / eyy1
                * (
                    -(k ** 2) * eyx1
                    - (1 - eyy1 / ezz1) / n / w
                    - eyx1 / ezz1 * (-2.0 / w ** 2 + 2 / w ** 2 * (e - w) / e)
                )
                + (ezz1 - ezz2) * (e - w) / e / w
                + 0.5
                * s
                * ezz2
                * ezz1
                / eyy2
                * (
                    -(k ** 2) * eyx2
                    + (1 - eyy2 / ezz2) / s / w
                    - eyx2 / ezz2 * (-2.0 / w ** 2 + 2 / w ** 2 * (e - w) / e)
                )
            )
            + (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
            * (
                0.5
                * n
                * ezz4
                * ezz3
                / eyy4
                * (
                    -(k ** 2) * eyx4
                    + (1 - eyy4 / ezz4) / n / e
                    - eyx4 / ezz4 * (-2.0 / e ** 2 - 2 / e ** 2 * (e - w) / w)
                )
                + 0.5
                * s
                * ezz3
                * ezz4
                / eyy3
                * (
                    -(k ** 2) * eyx3
                    - (1 - eyy3 / ezz3) / s / e
                    - eyx3 / ezz3 * (-2.0 / e ** 2 - 2 / e ** 2 * (e - w) / w)
                )
                + (ezz4 - ezz3) * (e - w) / e / w
            )
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
            * (
                0.5
                * ezz4
                * (-2.0 / n ** 2 - 2 * exx1 / ezz1 / w ** 2 + k ** 2 * exx1 - exy1 / ezz1 / n / w)
                / exx1
                * ezz1
                * w
                + 0.5
                * ezz1
                * (-2.0 / n ** 2 - 2 * exx4 / ezz4 / e ** 2 + k ** 2 * exx4 + exy4 / ezz4 / n / e)
                / exx4
                * ezz4
                * e
            )
            - (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
            * (
                0.5
                * ezz3
                * (-2.0 / s ** 2 - 2 * exx2 / ezz2 / w ** 2 + k ** 2 * exx2 + exy2 / ezz2 / s / w)
                / exx2
                * ezz2
                * w
                + 0.5
                * ezz2
                * (-2.0 / s ** 2 - 2 * exx3 / ezz3 / e ** 2 + k ** 2 * exx3 - exy3 / ezz3 / s / e)
                / exx3
                * ezz3
                * e
            )
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    ii = np.arange(nx * ny).reshape(nx, ny)

    # NORTH boundary

    ib = ii[:, -1]

    if boundary[0] == "S":
        sign = 1
    elif boundary[0] == "A":
        sign = -1
    elif boundary[0] == "0":
        sign = 0
    else:
        raise ValueError("unknown boundary conditions")

    bzxs[ib] += sign * bzxn[ib]
    bzxse[ib] += sign * bzxne[ib]
    bzxsw[ib] += sign * bzxnw[ib]
    bzys[ib] -= sign * bzyn[ib]
    bzyse[ib] -= sign * bzyne[ib]
    bzysw[ib] -= sign * bzynw[ib]

    # SOUTH boundary

    ib = ii[:, 0]

    if boundary[1] == "S":
        sign = 1
    elif boundary[1] == "A":
        sign = -1
    elif boundary[1] == "0":
        sign = 0
    else:
        raise ValueError("unknown boundary conditions")

    bzxn[ib] += sign * bzxs[ib]
    bzxne[ib] += sign * bzxse[ib]
    bzxnw[ib] += sign * bzxsw[ib]
    bzyn[ib] -= sign * bzys[ib]
    bzyne[ib] -= sign * bzyse[ib]
    bzynw[ib] -= sign * bzysw[ib]

    # EAST boundary

    ib = ii[-1, :]

    if boundary[2] == "S":
        sign = 1
    elif boundary[2] == "A":
        sign = -1
    elif boundary[2] == "0":
        sign = 0
    else:
        raise ValueError("unknown boundary conditions")

    bzxw[ib] += sign * bzxe[ib]
    bzxnw[ib] += sign * bzxne[ib]
    bzxsw[ib] += sign * bzxse[ib]
    bzyw[ib] -= sign * bzye[ib]
    bzynw[ib] -= sign * bzyne[ib]
    bzysw[ib] -= sign * bzyse[ib]

    # WEST boundary

    ib = ii[0, :]

    if boundary[3] == "S":
        sign = 1
    elif boundary[3] == "A":
        sign = -1
    elif boundary[3] == "0":
        sign = 0
    else:
        raise ValueError("unknown boundary conditions")

    bzxe[ib] += sign * bzxw[ib]
    bzxne[ib] += sign * bzxnw[ib]
    bzxse[ib] += sign * bzxsw[ib]
    bzye[ib] -= sign * bzyw[ib]
    bzyne[ib] -= sign * bzynw[ib]
    bzyse[ib] -= sign * bzysw[ib]

    # Assemble sparse matrix

    iall = ii.flatten()
    i_s = ii[:, :-1].flatten()
    i_n = ii[:, 1:].flatten()
    i_e = ii[1:, :].flatten()
    i_w = ii[:-1, :].flatten()
    i_ne = ii[1:, 1:].flatten()
    i_se = ii[1:, :-1].flatten()
    i_sw = ii[:-1, :-1].flatten()
    i_nw = ii[:-1, 1:].flatten()

    Izx = np.r_[iall, i_w, i_e, i_s, i_n, i_ne, i_se, i_sw, i_nw]
    Jzx = np.r_[iall, i_e, i_w, i_n, i_s, i_sw, i_nw, i_ne, i_se]
    Vzx = np.r_[
        bzxp[iall], bzxe[i_w], bzxw[i_e], bzxn[i_s], bzxs[i_n], bzxsw[i_ne], bzxnw[i_se], bzxne[i_sw], bzxse[i_nw]
    ]

    Izy = np.r_[iall, i_w, i_e, i_s, i_n, i_ne, i_se, i_sw, i_nw]
    Jzy = np.r_[iall, i_e, i_w, i_n, i_s, i_sw, i_nw, i_ne, i_se] + nx * ny
    Vzy = np.r_[
        bzyp[iall], bzye[i_w], bzyw[i_e], bzyn[i_s], bzys[i_n], bzysw[i_ne], bzynw[i_se], bzyne[i_sw], bzyse[i_nw]
    ]

    I = np.r_[Izx, Izy]
    J = np.r_[Jzx, Jzy]
    V = np.r_[Vzx, Vzy]
    B = coo_matrix((V, (I, J))).tocsr()

    HxHy = np.r_[Hx, Hy]
    Hz = B * HxHy.ravel() / 1j
    Hz = Hz.reshape(Hx.shape)

    # in xc e yc
    exx = epsxx[1:-1, 1:-1]
    exy = epsxy[1:-1, 1:-1]
    eyx = epsyx[1:-1, 1:-1]
    eyy = epsyy[1:-1, 1:-1]
    ezz = epszz[1:-1, 1:-1]
    edet = exx * eyy - exy * eyx

    h = e.reshape(nx, ny)[:-1, :-1]
    v = n.reshape(nx, ny)[:-1, :-1]

    # in xc e yc
    Dx = neff * EMpy_gpu.utils.centered2d(Hy) + (Hz[:-1, 1:] + Hz[1:, 1:] - Hz[:-1, :-1] - Hz[1:, :-1]) / (2j * k * v)
    Dy = -neff * EMpy_gpu.utils.centered2d(Hx) - (Hz[1:, :-1] + Hz[1:, 1:] - Hz[:-1, 1:] - Hz[:-1, :-1]) / (2j * k * h)
    Dz = (
        (Hy[1:, :-1] + Hy[1:, 1:] - Hy[:-1, 1:] - Hy[:-1, :-1]) / (2 * h)
        - (Hx[:-1, 1:] + Hx[1:, 1:] - Hx[:-1, :-1] - Hx[1:, :-1]) / (2 * v)
    ) / (1j * k)

    Ex = (eyy * Dx - exy * Dy) / edet
    Ey = (exx * Dy - eyx * Dx) / edet
    Ez = Dz / ezz

    return [Hx, Hy, Hz, Ex, Ey, Ez]



import numpy as np
from matplotlib import pyplot as plt
import pickle
import random
from copy import deepcopy
from typing import Callable


class EigenMode(object):
    """Virtual class representing an eigenmode"""

    def __init__(self):
        return NotImplementedError()

    def _inner_product(self):
        return NotImplementedError()

    def plot(self):
        return NotImplementedError()

    def get_confined_power(self):
        return NotImplementedError()

    def zero_phase(self):
        return NotImplementedError()

    def plot_material(self):
        return NotImplementedError()

    def save(self, path=("./ModeObject_" + str(random.random()) + ".pk")):
        """Serializes the mode into a pickle file"""
        pickle.dump(self, open(path, "wb+"))

    def get_fields(self):
        """Returns an array [self.Hx, self.Hy, self.Hz, self.Ex, self.Ey, self.Ez]."""

        return [self.Hx, self.Hy, self.Hz, self.Ex, self.Ey, self.Ez]

    def get_H(self):
        """Returns an array [self.Hx, self.Hy, self.Hz]."""

        return [self.Hx, self.Hy, self.Hz]

    def get_E(self):
        """Returns an array [self.Ex, self.Ey, self.Ez]."""

        return [self.Ex, self.Ey, self.Ez]

    def get_neff(self):
        """Returns the effective index as a complex number."""

        return self.neff

    def get_Hx(self):
        return self.Hx

    def get_Hy(self):
        return self.Hy

    def get_Hz(self):
        return self.Hz

    def get_Ex(self):
        return self.Ex

    def get_Ey(self):
        return self.Ey

    def get_Ez(self):
        return self.Ez

    def get_wavelength(self):
        """Returns the wavelength."""

        return self.wl

    def __str__(self):
        return "Mode Object with effective index of " + str(self.neff)

    def change_fields(self, start, other, func):
        for comp in ["Hx", "Hy", "Hz", "Ex", "Ey", "Ez"]:
            cur = getattr(start, comp)
            new = getattr(other, comp) if isinstance(other, EigenMode) else other
            setattr(start, comp, func(cur, new))
        return start

    def __mul__(self, other):
        return self.change_fields(deepcopy(self), other, lambda a, b: a * b)

    def __add__(self, other):
        return self.change_fields(deepcopy(self), other, lambda a, b: a + b)

    def __truediv__(self, other):
        return self.change_fields(deepcopy(self), other, lambda a, b: a / b)

    def __sub__(self, other):
        return self.change_fields(deepcopy(self), other, lambda a, b: a - b)

    def __imul__(self, other):
        return self.change_fields(self, other, lambda a, b: a * b)

    def __iadd__(self, other):
        return self.change_fields(self, other, lambda a, b: a + b)

    def __itruediv__(self, other):
        return self.change_fields(self, other, lambda a, b: a / b)

    def __isub__(self, other):
        return self.change_fields(self, other, lambda a, b: a - b)

    def inner_product(self, mode2) -> float:
        """Takes the inner product between self and the provided Mode

        Parameters
        ----------
        mode2 : EigenMode
            second eigenmode in the operation

        Returns
        -------
        number
            the inner product between the modes
        """

        return self._inner_product(self, mode2)

    def check_spurious(
        self, threshold_power: float = 0.05, threshold_neff: float = 0.9
    ) -> bool:
        """Takes in a mode and determine whether the mode is likely spurious based on the ratio of confined to not confined power

        Parameters
        ----------
        threshold_power : float
            threshold of power percentage of Pz in the core to total
        threshold_neff : float
            threshold of real to abs neff percentage

        Returns
        -------
        boolean
            True if likely spurious
        """

        power_bool = self.get_confined_power() < threshold_power
        neff_bool = (np.real(self.neff) / np.abs(self.neff)) < threshold_neff
        return power_bool or neff_bool

    def normalize(self) -> None:
        """Normalizes the Mode to power 1."""
        self.zero_phase()
        factor = self.inner_product(self)
        self /= np.sqrt(factor)


class Mode1D(EigenMode):
    """Object that holds the field profiles and effective index for a 1D eigenmode"""

    def __init__(
        self,
        x: "np.ndarray" = None,
        wl: float = None,
        neff: float = None,
        Hx: "np.ndarray" = None,
        Hy: "np.ndarray" = None,
        Hz: "np.ndarray" = None,
        Ex: "np.ndarray" = None,
        Ey: "np.ndarray" = None,
        Ez: "np.ndarray" = None,
        n: "np.ndarray" = None,
    ) -> None:
        """Constructor for Mode1D Object (one dimensional eigenmode)

        Parameters
        ----------
        x : (ndarray float)
            array of grid points in x direction (propogation in z)
        wl : (float)
            wavelength (meters)
        neff : (float)
            effective index
        Hx : (ndarray float)
            Hx field profile
        Hy : (ndarray float)
            Hy field profile
        Hz : (ndarray float)
            Hz field profile
        Ex : (ndarray float)
            Ex field profile
        Ey : (ndarray float)
            Ey field profile
        Ez : (ndarray float)
            Ez field profile
        n : (ndarray float)
            refractive index profile
        """

        self.x = x
        self.wl = wl
        self.neff = neff
        self.Hx = Hx if Hx is not None else np.zeros(10)
        self.Hy = Hy if Hy is not None else self.Hx * 0
        self.Hz = Hz if Hz is not None else self.Hx * 0
        self.Ex = Ex if Ex is not None else self.Hx * 0
        self.Ey = Ey if Ey is not None else self.Hx * 0
        self.Ez = Ez if Ez is not None else self.Hx * 0
        self.n = n
        self.H = np.sqrt(
            np.abs(self.Hx) ** 2 + np.abs(self.Hy) ** 2 + np.abs(self.Hz) ** 2
        )
        self.E = np.sqrt(
            np.abs(self.Ex) ** 2 + np.abs(self.Ey) ** 2 + np.abs(self.Ez) ** 2
        )

    def plot(self, operation: str = "Real", normalize: bool = True) -> None:
        """Plots the fields in the mode using pyplot. Should call plt.figure() before and plt.show() or plt.savefig() after

        Parameters
        ----------
        operation : string or function
            the operation to perform on the fields from ("Real", "Imaginary", "Abs", "Abs^2") (default:"Real") or a function such as np.abs
        normalize : bool
            if true, will normalize biggest field to 1
        """

        temp = (
            self
            / max(
                [
                    np.abs(np.real(np.amax(i)))
                    for i in [self.Ex, self.Ey, self.Ez, self.Hx, self.Hy, self.Hz]
                ]
            )
            if normalize
            else self / 1
        )

        # Parse operation
        op_name = (
            operation.__name__ if hasattr(operation, "__name__") else str(operation)
        )
        if operation == "Imaginary":
            operation = lambda a: np.imag(a)
        elif operation == "Abs":
            operation = lambda a: np.abs(a)
        elif operation == "Abs^2":
            operation = lambda a: np.abs(a) ** 2
        elif operation == "Real":
            operation = lambda a: np.real(a)
        try:
            t = self.change_fields(
                deepcopy(temp), deepcopy(temp), lambda a, b: operation(b)
            )
            Hx, Hy, Hz, Ex, Ey, Ez = [t.Hx, t.Hy, t.Hz, t.Ex, t.Ey, t.Ez]
        except Exception as e:
            print(e)
            raise Exception(
                "Invalid operation provided. Please choose from ('Imaginary', 'Abs', 'Abs^2', 'Real') or provide a function"
            )

        # Plot fields
        fields = ["Hx", "Hy", "Hz", "Ex", "Ey", "Ez"]
        for i, field in enumerate([Hx, Hy, Hz, Ex, Ey, Ez]):
            plt.subplot(2, 3, i + 1)
            plt.plot(self.x, field)
            plt.xlabel("x µm")
            plt.ylabel("{}({})".format(op_name, fields[i]))
        plt.tight_layout()

    def _inner_product(
        self, mode1: EigenMode, mode2: EigenMode, mask: "np.ndarray" = None
    ) -> float:
        """Helper function that takes the inner product between Modes mode1 and mode2

        Parameters
        ----------
        mode1 : Mode
            first eigenmode in the operation
        mode2 : Mode
            second eigenmode in the operation
        mask : np.ndarray
            a mask to multiply on the field profiles before conducting the inner product

        Returns
        -------
        number
            the inner product between the two input modes
        """

        mask = 1 if mask is None else mask

        Ex = mode1.Ex * mask
        Hy = np.conj(mode2.Hy) * mask
        Ey = mode1.Ey * mask
        Hx = np.conj(mode2.Hx) * mask

        cross = Ex * Hy - Ey * Hx

        return np.trapz(cross, np.real(mode1.x))

    def get_confined_power(self, num_pixels: int = None) -> float:
        """Takes in a mode and returns the percentage of power confined in the core

        Parameters
        ----------
        num_pixels : int
            number of pixels outside of the core to expand the mask to capture power just outside the core

        Returns
        -------
        float
            Percentage of confined power
        """

        # Increase core by 5% to capture slight leaks
        if num_pixels is None:
            num_pixels = int(len(self.x) * 0.05)

        mask = np.where(self.n > np.mean(self.n), 1, 0)
        kernel = np.ones(num_pixels + 1)
        mask = np.convolve(mask, kernel, "same")
        mask = np.where(mask > 0, 1, 0)
        ratio = self._inner_product(self, self, mask=mask) / self._inner_product(
            self, self, mask=None
        )
        return ratio

    def zero_phase(self) -> None:
        """Changes the phase such that the z components are all imaginary and the xy components are all real."""

        index = int(self.Hy.shape[0] / 2)
        phase = np.angle(np.array(self.Hy))[index]
        self *= np.exp(-1j * phase)
        if (np.sum(np.real(self.Hy))) < 0:
            self *= -1

    def plot_material(
        self, operation: Callable[["np.ndarray"], "np.ndarray"] = np.real
    ) -> None:
        """Plots the index of refraction profile"""
        plt.plot(self.x, operation(self.n))
        plt.title("Index of Refraction")
        plt.xlabel("x (µm)")
        plt.ylabel("y (µm)")


class Mode(EigenMode):
    """Object that holds the field profiles and effective index for a 2D eigenmode"""

    def __init__(
        self,
        x: "np.ndarray" = None,
        y: "np.ndarray" = None,
        wl: float = None,
        neff: float = None,
        Hx: "np.ndarray" = None,
        Hy: "np.ndarray" = None,
        Hz: "np.ndarray" = None,
        Ex: "np.ndarray" = None,
        Ey: "np.ndarray" = None,
        Ez: "np.ndarray" = None,
        n: "np.ndarray" = None,
    ) -> None:
        """Constructor for Mode Object

        Parameters
        ----------
        x : (ndarray float)
            array of grid points in x direction (propogation in z)
        y : (ndarray float)
            array of grid points in y direction (propogation in z)
        wl : (float)
            wavelength (meters)
        neff : (float)
            effective index
        Hx : (ndarray float)
            Hx field profile
        Hy : (ndarray float)
            Hy field profile
        Hz : (ndarray float)
            Hz field profile
        Ex : (ndarray float)
            Ex field profile
        Ey : (ndarray float)
            Ey field profile
        Ez : (ndarray float)
            Ez field profile
        n : (ndarray float)
            refractive index profile
        """

        self.x = x
        self.y = y
        self.wl = wl
        self.neff = neff
        self.Hx = Hx if Hx is not None else np.zeros((10, 10))
        self.Hy = Hy if Hy is not None else self.Hx * 0
        self.Hz = Hz if Hz is not None else self.Hx * 0
        self.Ex = Ex if Ex is not None else self.Hx * 0
        self.Ey = Ey if Ey is not None else self.Hx * 0
        self.Ez = Ez if Ez is not None else self.Hx * 0
        self.n = n
        self.H = np.sqrt(
            np.abs(self.Hx) ** 2 + np.abs(self.Hy) ** 2 + np.abs(self.Hz) ** 2
        )
        self.E = np.sqrt(
            np.abs(self.Ex) ** 2 + np.abs(self.Ey) ** 2 + np.abs(self.Ez) ** 2
        )

    def plot(
        self, operation: str = "Real", colorbar: bool = True, normalize: bool = True
    ) -> None:
        """Plots the fields in the mode using pyplot. Should call plt.figure() before and plt.show() or plt.savefig() after

        Parameters
        ----------
        operation : string or function
            the operation to perform on the fields from ("Real", "Imaginary", "Abs", "Abs^2") (default:"Real") or a function such as np.abs
        colorbar : bool
            if true, will show a colorbar for each field
        normalize : bool
            if true, will normalize biggest field to 1
        """

        temp = (
            self
            / max(
                [
                    np.abs(np.real(np.amax(i)))
                    for i in [self.Ex, self.Ey, self.Ez, self.Hx, self.Hy, self.Hz]
                ]
            )
            if normalize
            else self / 1
        )

        # Parse operation
        op_name = (
            operation.__name__ if hasattr(operation, "__name__") else str(operation)
        )
        if operation == "Imaginary":
            operation = lambda a: np.imag(a)
        elif operation == "Abs":
            operation = lambda a: np.abs(a)
        elif operation == "Abs^2":
            operation = lambda a: np.abs(a) ** 2
        elif operation == "Real":
            operation = lambda a: np.real(a)
        try:
            t = self.change_fields(
                deepcopy(temp), deepcopy(temp), lambda a, b: operation(b)
            )
            Hx, Hy, Hz, Ex, Ey, Ez = [t.Hx, t.Hy, t.Hz, t.Ex, t.Ey, t.Ez]
        except Exception as e:
            print(e)
            raise Exception(
                "Invalid operation provided. Please choose from ('Imaginary', 'Abs', 'Abs^2', 'Real') or provide a function"
            )

        # Plot fields
        fields = ["Hx", "Hy", "Hz", "Ex", "Ey", "Ez"]
        for i, field in enumerate([Hx, Hy, Hz, Ex, Ey, Ez]):
            plt.subplot(
                2, 3, i + 1, adjustable="box", aspect=field.shape[0] / field.shape[1]
            )
            v = max(abs(field.min()), abs(field.max()))
            plt.imshow(
                np.rot90(field),
                cmap="RdBu",
                vmin=-v,
                vmax=v,
                extent=[
                    temp.x[0],
                    temp.x[-1],
                    temp.y[0],
                    temp.y[-1],
                ],
                interpolation="none",
            )
            plt.title("{}({})".format(op_name, fields[i]))
            if colorbar:
                plt.colorbar(fraction=0.046, pad=0.04)
            plt.xlabel("x µm")
            plt.ylabel("y µm")

        plt.tight_layout()

    def _inner_product(
        self, mode1: EigenMode, mode2: EigenMode, mask: "np.ndarray" = None
    ) -> float:
        """Helper function that takes the inner product between Modes mode1 and mode2

        Parameters
        ----------
        mode1 : EigenMode
            first eigenmode in the operation
        mode2 : EigenMode
            second eigenmode in the operation
        mask : np.ndarray
            a mask to multiply on the field profiles before conducting the inner product
        Returns
        -------
        number
            the inner product between the two input modes
        """

        mask = 1 if mask is None else mask

        Ex = mode1.Ex * mask
        Hy = np.conj(mode2.Hy) * mask
        Ey = mode1.Ey * mask
        Hx = np.conj(mode2.Hx) * mask

        cross = Ex * Hy - Ey * Hx

        return np.trapz(np.trapz(cross, np.real(mode1.x)), np.real(mode1.y))

    def get_confined_power(self, num_pixels: int = None) -> float:
        """Takes in a mode and returns the percentage of power confined in the core

        Parameters
        ----------
        num_pixels : int
            number of pixels outside of the core to expand the mask to capture power just outside the core (mask dilation)

        Returns
        -------
        float
            Percentage of confined power
        """

        # Increase core by 5% to capture slight leaks
        if num_pixels is None:
            num_pixels = int(len(self.x) * 0.05)

        mask = np.where(self.n > np.mean(self.n), 1, 0)
        kernel = np.ones(num_pixels + 1)
        for row in range(mask.shape[0]): # This could be vectorized with scipy or just a 2D kernal, but numpy won't let this happen
            mask[row] = np.convolve(mask[row], kernel, "same")
        for col in range(mask.shape[1]):
            mask[:, col] = np.convolve(mask[:, col], kernel, "same")
        mask = np.where(mask > 0, 1, 0)
        ratio = self._inner_product(self, self, mask=mask) / self._inner_product(
            self, self, mask=None
        )
        return ratio

    def zero_phase(self) -> None:
        """Changes the phase such that the z components are all imaginary and the xy components are all real."""

        index = int(self.Hy.shape[0] / 2)
        phase = np.angle(np.array(self.Hy))[index][index]
        self *= np.exp(-1j * phase)
        if (np.sum(np.real(self.Hy))) < 0:
            self *= -1

    def plot_material(self) -> None:
        """Plots the index of refraction profile"""

        plt.figure()
        plt.imshow(
            np.real(np.rot90(self.n)),
            extent=[
                self.x[0],
                self.x[-1],
                self.y[0],
                self.y[-1],
            ],
            cmap="Greys",
            interpolation="none",
        )
        plt.colorbar()
        plt.title("Index of Refraction")
        plt.xlabel("x (µm)")
        plt.ylabel("y (µm)")



import numpy as np
from matplotlib import pyplot as plt
import EMpy_gpu
from EMpy_gpu.modesolvers.FD import stretchmesh
from typing import Callable

from emepy.mode import Mode, Mode1D, EigenMode
from emepy.tools import interp, interp1d, Si, SiO2, get_epsfunc, rectangle_to_n


class ModeSolver(object):
    """The ModeSolver object is the heart of finding eigenmodes for use in eigenmode expansion or simple examination. This parent class should be inherited and used as a wrapper for certain modules such as EMpy, Lumerical, Pickled data, Neural Networks, etc."""

    def __init__(self, **kwargs) -> None:
        """ModeSolver class constructor"""
        raise NotImplementedError

    def solve(self) -> None:
        """Solves the eigenmode solver for the specific eigenmodes of desire"""
        raise NotImplementedError

    def clear(self) -> None:
        """Clears the modesolver's eigenmodes to make memory"""
        raise NotImplementedError

    def get_mode(self, mode_num: int) -> EigenMode:
        """Must extract the mode of choice

        Parameters
        ----------
        mode_num : int
            index of the mode of choice
        """
        raise NotImplementedError


class ModeSolver1D(ModeSolver):
    pass


class MSEMpy(ModeSolver):
    """Electromagnetic Python Modesolver. Uses the EMpy library See Modesolver. Parameterizes the cross section as a rectangular waveguide."""

    def __init__(
        self,
        wl: float = 1.55,
        width: float = None,
        thickness: float = None,
        num_modes: int = 1,
        cladding_width: float = 2.5,
        cladding_thickness: float = 2.5,
        core_index: float = None,
        cladding_index: float = None,
        x: "np.ndarray" = None,
        y: "np.ndarray" = None,
        mesh: int = 128,
        accuracy: float = 1e-8,
        boundary: str = "0000",
        epsfunc: Callable[["np.ndarray", "np.ndarray"], "np.ndarray"] = None,
        n: "np.ndarray" = None,
        PML: bool = False,
        subpixel: bool = True,
        center: tuple = (0, 0),
        **kwargs
    ) -> None:
        """MSEMpy class constructor

        Parameters
        ----------
        wl : number
            wavelength of the eigenmodes
        width : number
            width of the core in the cross section
        thickness : number
            thickness of the core in the cross section
        num_modes : int
            number of modes to solve for (default:1)
        cladding_width : number
            width of the cladding in the cross section (default:5)
        cladding_thickness : number
            thickness of the cladding in the cross section (default:5)
        core_index : number
            refractive index of the core (default:Si)
        cladding_index : number
            refractive index of the cladding (default:SiO2)
        mesh : int
            number of mesh points in each direction (xy)
        x : numpy array
            the cross section grid in the x direction (z propagation) (default:None)
        y : numpy array
            the cross section grid in the y direction (z propagation) (default:None)
        mesh : int
            the number of mesh points in each xy direction
        accuracy : number
            the minimum accuracy of the finite difference solution (default:1e-8)
        boundary : string
            the boundaries according to the EMpy library (default:"0000")
        epsfunc : function
            the function which defines the permittivity based on a grid (see EMpy library) (default:"0000")
        n : numpy array
            2D profile of the refractive index
        PML : bool
            if True, will use PML boundaries. Only works for Tidy3D, not EMpy. Default : False, PEC
        subpixel : bool
            if true, will use subpixel smoothing, assuming asking for a waveguide cross section and not providing an index map (recommended)
        """

        self.wl = wl
        self.width = width
        self.thickness = thickness
        self.num_modes = num_modes
        self.cladding_width = cladding_width
        self.cladding_thickness = cladding_thickness
        self.core_index = core_index if core_index is not None else Si(wl)
        self.cladding_index = cladding_index if cladding_index is not None else SiO2(wl)
        self.x = (
            x
            if x is not None
            else np.linspace(-0.5 * cladding_width, 0.5 * cladding_width, mesh)
        )
        self.y = (
            y
            if y is not None
            else np.linspace(-0.5 * cladding_width, 0.5 * cladding_width, mesh)
        )
        self.mesh = mesh
        self.accuracy = accuracy
        self.boundary = boundary
        self.epsfunc = epsfunc
        self.n = n
        self.PML = PML

        if epsfunc is None and (not subpixel) or (self.width is None):
            self.epsfunc = get_epsfunc(
                self.width,
                self.thickness,
                self.cladding_width,
                self.cladding_thickness,
                self.core_index,
                self.cladding_index,
                profile=self.n,
                nx=self.x,
                ny=self.y,
            )
        elif epsfunc is None and subpixel and (self.width is not None):
            n = rectangle_to_n(
                center,
                self.width,
                self.thickness,
                self.x,
                self.y,
                subpixel,
                self.core_index,
                self.cladding_index,
            )
            self.x = ((self.x)[1:] + (self.x)[:-1]) / 2
            self.y = ((self.y)[1:] + (self.y)[:-1]) / 2
            self.epsfunc = get_epsfunc(
                None,
                None,
                self.cladding_width,
                self.cladding_thickness,
                self.core_index,
                self.cladding_index,
                profile=n,
                nx=self.x,
                ny=self.y,
            )

        self.after_x = self.x
        self.after_y = self.y
        self.n = np.sqrt(self.epsfunc(self.x, self.y))

    def solve(self) -> ModeSolver:
        """Solves for the eigenmodes"""
        self.solver = EMpy_gpu.modesolvers.FD.VFDModeSolver(
            self.wl, self.x, self.y, self.epsfunc, self.boundary
        ).solve(self.num_modes, self.accuracy)
        return self

    def clear(self) -> ModeSolver:
        """Clears the modesolver's eigenmodes to make memory"""

        self.solver = None
        return self

    def get_mode(self, mode_num: int = 0) -> EigenMode:
        """Get the indexed mode number

        Parameters
        ----------
        mode_num : int
            index of the mode of choice

        Returns
        -------
        Mode
            the eigenmode of index mode_num
        """
        x = self.solver.modes[mode_num].get_x()
        y = self.solver.modes[mode_num].get_y()
        x0, y0 = [np.real(x), np.real(y)]
        diffx, diffy = [np.diff(x0), np.diff(y0)]
        x0_new, y0_new = [np.ones(len(x) + 1), np.ones(len(y) + 1)]
        x0_new[0:-1], y0_new[0:-1] = [x0, y0]
        x0_new[-1], y0_new[-1] = [x0[-1] + diffx[-1], y0[-1] + diffy[-1]]

        # if not self.PML:
        self.nlayers = [1, 0, 1, 0]
        Ex = interp(
            x0_new, y0_new, x0, y0, self.solver.modes[mode_num].get_field("Ex"), True
        )[self.nlayers[1] : -self.nlayers[0], self.nlayers[3] : -self.nlayers[2]]
        Ey = interp(
            x0_new, y0_new, x0, y0, self.solver.modes[mode_num].get_field("Ey"), True
        )[self.nlayers[1] : -self.nlayers[0], self.nlayers[3] : -self.nlayers[2]]
        Ez = interp(
            x0_new, y0_new, x0, y0, self.solver.modes[mode_num].get_field("Ez"), True
        )[self.nlayers[1] : -self.nlayers[0], self.nlayers[3] : -self.nlayers[2]]
        Hx = interp(
            x0_new, y0_new, x0, y0, self.solver.modes[mode_num].get_field("Hx"), False
        )[self.nlayers[1] : -self.nlayers[0], self.nlayers[3] : -self.nlayers[2]]
        Hy = interp(
            x0_new, y0_new, x0, y0, self.solver.modes[mode_num].get_field("Hy"), False
        )[self.nlayers[1] : -self.nlayers[0], self.nlayers[3] : -self.nlayers[2]]
        Hz = interp(
            x0_new, y0_new, x0, y0, self.solver.modes[mode_num].get_field("Hz"), False
        )[self.nlayers[1] : -self.nlayers[0], self.nlayers[3] : -self.nlayers[2]]
        self.x = x0_new[self.nlayers[1] : -self.nlayers[0]]
        self.y = y0_new[self.nlayers[3] : -self.nlayers[2]]

        neff = self.solver.modes[mode_num].neff
        n = np.sqrt(self.epsfunc(self.x, self.y))

        return Mode(
            x=self.x,
            y=self.y,
            wl=self.wl,
            neff=neff,
            Hx=Hx,
            Hy=Hy,
            Hz=Hz,
            Ex=Ex,
            Ey=Ey,
            Ez=Ez,
            n=n,
        )

    def plot_material(self) -> None:
        """Plots the index of refraction profile"""
        plt.imshow(
            np.sqrt(np.real(self.n)).T,
            extent=[self.x[0], self.x[-1], self.y[0], self.y[-1]],
            cmap="Greys",
        )
        plt.colorbar()
        plt.title("Index of Refraction")
        plt.xlabel("x (µm)")
        plt.ylabel("y (µm)")


class MSTidy3D(MSEMpy):
    def solve(self) -> ModeSolver:
        """Solves for the eigenmodes"""
        from tidy3d.plugins.mode.solver import compute_modes
        import tidy3d as td
        from types import SimpleNamespace

        bound_x = 0.5 * (self.x[1:] + self.x[:-1])
        bound_x = np.insert(bound_x, 0, bound_x[0] - np.diff(bound_x)[0])
        bound_x = np.insert(bound_x, len(bound_x), bound_x[-1] + np.diff(bound_x)[-1])

        bound_y = 0.5 * (self.y[1:] + self.y[:-1])
        bound_y = np.insert(bound_y, 0, bound_y[0] - np.diff(bound_y)[0])
        bound_y = np.insert(bound_y, len(bound_y), bound_y[-1] + np.diff(bound_y)[-1])

        (E, H), neff = compute_modes(
            eps_cross=np.array([self.n**2, self.n**2, self.n**2]),
            coords=(bound_x, bound_y),
            freq=td.constants.C_0 / (self.wl),
            mode_spec=SimpleNamespace(
                num_modes=int(self.num_modes),
                bend_radius=None,
                bend_axis=1,
                angle_theta=0.0,
                angle_phi=0.0,
                num_pml=(0, 0) if not self.PML else (20, 20),
                target_neff=self.core_index,
                precision="single",
                filter_pol=None,
            ),
        )

        self.modes = [
            Mode(
                self.x,
                self.y,
                self.wl,
                neff[i],
                H[0, :, :, 0, i],
                H[1, :, :, 0, i],
                H[2, :, :, 0, i],
                E[0, :, :, 0, i],
                E[1, :, :, 0, i],
                E[2, :, :, 0, i],
                self.n,
            )
            for i in range(self.num_modes)
        ]

        return self

    def clear(self) -> ModeSolver:
        """Clears the modesolver's eigenmodes to make memory"""

        self.solver = None
        return self

    def get_mode(self, mode_num: int = 0) -> EigenMode:
        return self.modes[mode_num]


# class MSEMpy1D(ModeSolver):
#     """NOTICE: DOES NOT CURRENTLY WORK!! Electromagnetic Python Modesolver. Uses the EMpy library See Modesolver. Parameterizes the cross section as a rectangular waveguide."""

#     def __init__(
#         self,
#         wl: float,
#         width: float = None,
#         num_modes: int = 1,
#         cladding_width: float = 2.5,
#         core_index: float = None,
#         cladding_index: float = None,
#         x: "np.ndarray" = None,
#         mesh: int = 128,
#         accuracy: float = 1e-8,
#         boundary: str = "0000",
#         epsfunc: Callable[["np.ndarray", "np.ndarray"], "np.ndarray"] = None,
#         n: "np.ndarray" = None,
#         PML: bool = False,
#         **kwargs
#     ):
#         """MSEMpy class constructor

#         Parameters
#         ----------
#         wl : number
#             wavelength of the eigenmodes
#         width : number
#             width of the core in the cross section
#         num_modes : int
#             number of modes to solve for (default:1)
#         cladding_width : number
#             width of the cladding in the cross section (default:5)
#         core_index : number
#             refractive index of the core (default:Si)
#         cladding_index : number
#             refractive index of the cladding (default:SiO2)
#         mesh : int
#             number of mesh points in each direction (xy)
#         x : numpy array
#             the cross section grid in the x direction (z propagation) (default:None)
#         mesh : int
#             the number of mesh points in each xy direction
#         accuracy : number
#             the minimum accuracy of the finite difference solution (default:1e-8)
#         boundary : string
#             the boundaries according to the EMpy library (default:"0000")
#         epsfunc : function
#             the function which defines the permittivity based on a grid (see EMpy library) (default:"0000")
#         n : numpy array
#             2D profile of the refractive index
#         PML : boolean
#             if True, will use PML boundaries. Default : False, PEC
#         """

#         self.wl = wl
#         self.width = width
#         self.num_modes = num_modes
#         self.cladding_width = cladding_width
#         self.core_index = core_index
#         self.cladding_index = cladding_index
#         self.x = x
#         self.mesh = mesh
#         self.accuracy = accuracy
#         self.boundary = boundary
#         self.epsfunc = epsfunc
#         self.n = n
#         self.PML = PML

#         if core_index is None:
#             self.core_index = Si(wl)
#         if cladding_index is None:
#             self.cladding_index = SiO2(wl)
#         if x is None:
#             self.x = np.linspace(-0.5 * cladding_width, 0.5 * cladding_width, mesh)
#         if self.PML:  # Create a PML at least half a wavelength long
#             dx = np.diff(self.x)
#             layer_xp = int(np.abs(0.5 * self.wl / dx[-1]))
#             layer_xn = int(np.abs(0.5 * self.wl / dx[0]))
#             self.nlayers = [layer_xp, layer_xn, 0, 0]
#             factor = 1 + 2j
#             self.x, _, _, _, _, _ = stretchmesh(self.x, np.zeros(1), self.nlayers, factor)
#         if epsfunc is None:
#             self.epsfunc = get_epsfunc(
#                 self.width,
#                 None,
#                 self.cladding_width,
#                 None,
#                 self.core_index,
#                 self.cladding_index,
#                 profile=self.n,
#                 nx=self.x,
#             )

#         self.after_x = self.x
#         self.n = self.epsfunc(self.x, np.zeros(1))

#     def solve(self) -> ModeSolver:
#         """Solves for the eigenmodes"""
#         self.solver = EMpy_gpu.modesolvers.FD.VFDModeSolver(
#             self.wl, self.x, np.zeros(1), self.epsfunc, self.boundary
#         ).solve(self.num_modes, self.accuracy)
#         return self

#     def clear(self) -> ModeSolver:
#         """Clears the modesolver's eigenmodes to make memory"""

#         self.solver = None
#         return self

#     def get_mode(self, mode_num: int = 0) -> EigenMode:
#         """Get the indexed mode number

#         Parameters
#         ----------
#         mode_num : int
#             index of the mode of choice

#         Returns
#         -------
#         Mode
#             the eigenmode of index mode_num
#         """

#         x = self.solver.modes[mode_num].get_x()
#         x0, y0 = [np.real(x), np.real(x)]
#         diffx, diffy = [np.diff(x0), np.diff(y0)]
#         x0_new, y0_new = [np.ones(len(x) + 1), np.ones(len(x) + 1)]
#         x0_new[0:-1], y0_new[0:-1] = [x0, y0]
#         x0_new[-1], y0_new[-1] = [x0[-1] + diffx[-1], y0[-1] + diffy[-1]]

#         if not self.PML:
#             self.nlayers = [1, 0, 1, 0]
#         Ex = interp1d(x0_new, x0.self.solver.modes[mode_num].get_field("Ex"), True)[self.nlayers[1] : -self.nlayers[0]]
#         Ey = interp1d(x0_new, x0.self.solver.modes[mode_num].get_field("Ey"), True)[self.nlayers[1] : -self.nlayers[0]]
#         Ez = interp1d(x0_new, x0.self.solver.modes[mode_num].get_field("Ez"), True)[self.nlayers[1] : -self.nlayers[0]]
#         Hx = interp1d(x0_new, x0.self.solver.modes[mode_num].get_field("Hx"), False)[self.nlayers[1] : -self.nlayers[0]]
#         Hy = interp1d(x0_new, x0.self.solver.modes[mode_num].get_field("Hy"), False)[self.nlayers[1] : -self.nlayers[0]]
#         Hz = interp1d(x0_new, x0.self.solver.modes[mode_num].get_field("Hz"), False)[self.nlayers[1] : -self.nlayers[0]]
#         self.x = x0_new[self.nlayers[1] : -self.nlayers[0]]

#         neff = self.solver.modes[mode_num].neff
#         n = np.sqrt(self.epsfunc(self.x, np.zeros(0)))

#         return Mode1D(x=self.x, wl=self.wl, neff=neff, Hx=Hx, Hy=Hy, Hz=Hz, Ex=Ex, Ey=Ey, Ez=Ez, n=n)

#     def plot_material(self) -> None:
#         """Plots the index of refraction profile"""
#         plt.plot(self.x, self.n)
#         plt.title("Index of Refraction")
#         plt.xlabel("x (µm)")
#         plt.ylabel("y (µm)")


# class MSPickle(object):
#     """Pickle Modesolver. See Modesolver. Pickle should serialize a list of Mode objects that can be opened here."""

#     def __init__(
#         self, filename: str, index: int = None, width: float = None, thickness: float = None, **kwargs
#     ) -> None:
#         """MSPickle class constructor

#         Parameters
#         ----------
#         filename : string
#             the name of the pickled file where the eigenmode is stored
#         index : int
#             the index of the mode in the pickle file if the data stored is an array of Modes (default:None)
#         width : number
#             width of the core in the cross section, used for drawing (default:None)
#         thickness : number
#             thickness of the core in the cross section, used for drawing  (default:None)
#         """

#         self.filename = filename
#         self.index = index
#         self.width = width
#         self.thickness = thickness
#         self.PML = False

#     def solve(self) -> ModeSolver:
#         """Solves for the eigenmodes by loading them from the pickle file"""

#         with open(self.filename, "rb") as f:
#             self.mode = pickle.load(f)[self.index] if self.index else pickle.load(f)

#         self.x = self.mode.x
#         self.y = self.mode.y
#         return self

#     def clear(self) -> ModeSolver:
#         """Clears the modesolver's eigenmodes to make memory"""

#         self.x = None
#         self.y = None
#         self.mode = None
#         return self

#     def get_mode(self, mode_num: int = 0) -> EigenMode:
#         """Get the stored mode

#         Returns
#         -------
#         Mode
#             the eigenmode of index mode_num
#         """

#         return self.mode



import numpy as np
from emepy.fd import MSEMpy, ModeSolver, MSTidy3D
from emepy.models import Layer
from emepy.tools import vertices_to_n
from copy import deepcopy

"""
Write polygon that utilizes tidy3d
"""


class Geometry(object):
    """Geoemtries are not required for users, however they do allow for easier creation of complex structures"""

    def __init__(self, layers: list) -> None:
        """Constructors should take in parameters from the user and build the layers"""
        self.layers = layers

    def __iter__(self):
        return deepcopy(iter(self.layers))


class Params(object):
    def __init__(self) -> None:
        return

    def get_solver_rect(self) -> ModeSolver:
        return

    def get_solver_index(self) -> ModeSolver:
        return


class EMpyGeometryParameters(Params):
    def __init__(
        self,
        wavelength: float = 1.55,
        cladding_width: float = 2.5,
        cladding_thickness: float = 2.5,
        core_index: float = None,
        cladding_index: float = None,
        x: "np.ndarray" = None,
        y: "np.ndarray" = None,
        mesh: int = 128,
        accuracy: float = 1e-8,
        boundary: str = "0000",
        PML: bool = False,
        **kwargs,
    ):
        """Creates an instance of EMpyGeometryParameters which is used for abstract geometries that use EMpy as the solver"""

        self.wavelength = wavelength
        self.cladding_width = cladding_width
        self.cladding_thickness = cladding_thickness
        self.core_index = core_index
        self.cladding_index = cladding_index
        self.x = x
        self.y = y
        self.mesh = mesh
        self.accuracy = accuracy
        self.boundary = boundary
        self.PML = PML
        for key, val in kwargs.items():
            setattr(key, val)

    def get_solver_rect(
        self,
        width: float = 0.5,
        thickness: float = 0.22,
        num_modes: int = 1,
        center: bool = (0, 0),
    ) -> "MSEMpy":
        """Returns an EMPy solver that represents a simple rectangle"""

        return MSEMpy(
            wl=self.wavelength,
            width=width,
            thickness=thickness,
            num_modes=num_modes,
            cladding_width=self.cladding_width,
            cladding_thickness=self.cladding_thickness,
            core_index=self.core_index,
            cladding_index=self.cladding_index,
            x=self.x,
            y=self.y,
            mesh=self.mesh,
            accuracy=self.accuracy,
            boundary=self.boundary,
            PML=self.PML,
            center=center,
        )

    def get_solver_index(
        self, thickness: float = None, num_modes: int = None, n: "np.ndarray" = None
    ) -> "MSEMpy":
        """Returns an EMPy solver that represents the provided index profile"""

        return MSEMpy(
            wl=self.wavelength,
            width=None,
            thickness=thickness,
            num_modes=num_modes,
            cladding_width=self.cladding_width,
            cladding_thickness=self.cladding_thickness,
            core_index=self.core_index,
            cladding_index=self.cladding_index,
            x=self.x,
            y=self.y,
            mesh=self.mesh,
            accuracy=self.accuracy,
            boundary=self.boundary,
            n=n,
            PML=self.PML,
        )


class MSTidy3DGeometryParameters(Params):
    def __init__(
        self,
        wavelength: float = 1.55,
        cladding_width: float = 2.5,
        cladding_thickness: float = 2.5,
        core_index: float = None,
        cladding_index: float = None,
        x: "np.ndarray" = None,
        y: "np.ndarray" = None,
        mesh: int = 128,
        accuracy: float = 1e-8,
        boundary: str = "0000",
        PML: bool = False,
        **kwargs,
    ):
        """Creates an instance of Tidy3DGeometryParameters which is used for abstract geometries that use Tidy3D as the solver"""

        self.wavelength = wavelength
        self.cladding_width = cladding_width
        self.cladding_thickness = cladding_thickness
        self.core_index = core_index
        self.cladding_index = cladding_index
        self.x = x
        self.y = y
        self.mesh = mesh
        self.accuracy = accuracy
        self.boundary = boundary
        self.PML = PML
        for key, val in kwargs.items():
            setattr(key, val)

    def get_solver_rect(
        self,
        width: float = 0.5,
        thickness: float = 0.22,
        num_modes: int = 1,
        center: bool = (0, 0),
    ) -> "MSTidy3D":
        """Returns an MSTidy3D solver that represents a simple rectangle"""

        return MSTidy3D(
            wl=self.wavelength,
            width=width,
            thickness=thickness,
            num_modes=num_modes,
            cladding_width=self.cladding_width,
            cladding_thickness=self.cladding_thickness,
            core_index=self.core_index,
            cladding_index=self.cladding_index,
            x=self.x,
            y=self.y,
            mesh=self.mesh,
            accuracy=self.accuracy,
            boundary=self.boundary,
            PML=self.PML,
            center=center,
        )

    def get_solver_index(
        self, thickness: float = None, num_modes: int = None, n: "np.ndarray" = None
    ) -> "MSTidy3D":
        """Returns an MSTidy3D solver that represents the provided index profile"""

        return MSTidy3D(
            wl=self.wavelength,
            width=None,
            thickness=thickness,
            num_modes=num_modes,
            cladding_width=self.cladding_width,
            cladding_thickness=self.cladding_thickness,
            core_index=self.core_index,
            cladding_index=self.cladding_index,
            x=self.x,
            y=self.y,
            mesh=self.mesh,
            accuracy=self.accuracy,
            boundary=self.boundary,
            n=n,
            PML=self.PML,
        )


class DynamicPolygon(Geometry):
    """Creates a polygon in EMEPy given a list of solid vertices and changeable vertices and can be changed for shape optimization"""

    def get_design(self) -> list:
        """Returns the design region as a list of parameters. Each parameter represents a spacial location in a single direction for a single vertex of the polygon. Note in 2D this would look like [x0,z0,x1,z1,...xn,zn] and in 3D [x0,y0,z0,x1,y1,z1,...xn,yn,zn]"""
        return self.design

    def set_design(self) -> list:
        """Sets the design region"""
        return

    def get_n(self, grid_x, grid_y, grid_z) -> "np.ndarray":
        """Returns the index profile for the grid provided"""
        return

    def __len__(self):
        return len(self.get_design())


class DynamicRect2D(DynamicPolygon):
    def __init__(
        self,
        params: Params = MSTidy3DGeometryParameters(),
        width: float = 0.5,
        thickness: float = 0.22,
        length: float = 1,
        num_modes: int = 1,
        num_params: int = 10,
        symmetry: bool = False,
        subpixel: bool = True,
        mesh_z: int = 10,
        input_width: float = None,
        output_width: float = None,
    ) -> None:
        """Creates an instance of DynamicPolygon2D"""
        input_width = input_width if input_width is not None else width
        output_width = output_width if output_width is not None else width
        self.num_modes = num_modes
        self.params = deepcopy(params)
        self.symmetry = symmetry
        self.subpixel = subpixel
        self.width, self.thickness, self.length = (width, thickness, length)
        self.num_params = num_params
        self.grid_x = (
            params.x
            if params.x is not None
            else np.linspace(
                -params.cladding_width / 2, params.cladding_width / 2, params.mesh
            )
        )
        self.grid_z = np.linspace(0, length, mesh_z)

        # Set left side static vertices
        x = [-input_width / 2, input_width / 2]
        z = [0, 0]
        self.static_vertices_left = list(zip(x, z))

        # Set top dynamic vertices
        x = np.linspace(
            input_width / 2, output_width / 2, num_params
        ).tolist()  # [width / 2] * num_params
        z = np.linspace(0, length, num_params + 2)[1:-1].tolist()
        dynamic_vertices_top = list(zip(x, z))

        # Set right side static vertices
        x = [output_width / 2, -output_width / 2]
        z = [length, length]
        self.static_vertices_right = list(zip(x, z))

        # Set bottom dynamic vertices
        x = np.linspace(
            -output_width / 2, -input_width / 2, num_params
        ).tolist()  # [-width / 2] * num_params
        z = np.linspace(0, length, num_params + 2)[1:-1][::-1].tolist()
        dynamic_vertices_bottom = list(zip(x, z))

        # Establish design
        design = (
            dynamic_vertices_top[:]
            if symmetry
            else dynamic_vertices_top + dynamic_vertices_bottom
        )
        design = [i for j in design for i in j]

        # Fix params
        self.params.x = (
            0.5 * (self.params.x[1:] + self.params.x[:-1])
            if self.params.x is not None
            else self.params.x
        )
        self.params.y = (
            0.5 * (self.params.y[1:] + self.params.y[:-1])
            if self.params.y is not None
            else self.params.y
        )
        self.params.mesh -= 1

        # Set design
        self.set_design(design)

    def set_design(self, design: list):
        """Sets the design region parameters"""
        self.design = design
        self.set_layers()

    def get_n(self, grid_x, grid_y, grid_z):
        """Will form the refractive index map given the current parameters"""
        # Create vertices
        vertices = []

        # Add static left vertices
        vertices += self.static_vertices_left

        # Add top design
        top = self.design if self.symmetry else self.design[: len(self.design) // 2]
        vertices += [(x, z) for x, z in zip(top[:-1:2], top[1::2])]

        # Add static right vertices
        vertices += self.static_vertices_right

        # Add bottom design
        bottom = self.design if self.symmetry else self.design[len(self.design) // 2 :]
        if self.symmetry:
            vertices += [
                (-x, z) for x, z in list(zip(bottom[:-1:2], bottom[1::2]))[::-1]
            ]
        else:
            vertices += [(x, z) for x, z in list(zip(bottom[:-1:2], bottom[1::2]))]

        # Form polygon
        polygon = vertices_to_n(
            vertices,
            grid_x,
            grid_z,
            self.subpixel,
            self.params.core_index,
            self.params.cladding_index,
        )

        # Extend in 3D
        if grid_y is not None:
            polygon = np.stack([polygon] * (len(grid_y) - 1), axis=1)
            x, y, z = (
                0.5 * (grid_x[1:] + grid_x[:-1]),
                0.5 * (grid_y[1:] + grid_y[:-1]),
                0.5 * (grid_z[1:] + grid_z[:-1]),
            )
            x_, y_, z_ = np.meshgrid(x, y, z, indexing="ij")
            n = np.ones(x_.shape) * self.params.cladding_index
            n = np.where(np.abs(y_) < self.thickness / 2, polygon, n)
            return n

        # Return polygon
        return polygon

    def set_layers(self):
        """Creates the layers needed for the geometry"""

        # Get n
        n = self.get_n(self.grid_x, None, self.grid_z)
        from matplotlib import pyplot as plt

        # Iterate through n and create layers
        self.layers = []
        diff_z = np.diff(self.grid_z)
        for i, dz in enumerate(diff_z):
            mode_solver = self.params.get_solver_index(0.22, self.num_modes, n[:, i])
            layer = Layer(mode_solver, self.num_modes, self.params.wavelength, dz)
            self.layers.append(layer)


class Waveguide(Geometry):
    """Block forms the simplest geometry in emepy, a single layer with a single waveguide defined"""

    def __init__(
        self,
        params: Params = EMpyGeometryParameters(),
        width: float = 0.5,
        thickness: float = 0.22,
        length: float = 1,
        num_modes: int = 1,
        center: tuple = (0, 0),
    ) -> None:
        """Creates an instance of block which can be called to access the required layers for solving

        Parameters
        ----------
        params : Params
            Geometry Parameters object containing large scale parameters
        width : number
            width of the core in the cross section
        thickness : number
            thickness of the core in the cross section
        length : number
            length of the structure
        num_modes : int
            number of modes to solve for (default:1)
        """

        self.length = length
        self.width = width
        mode_solver = params.get_solver_rect(width, thickness, num_modes, center=center)
        layers = [Layer(mode_solver, num_modes, params.wavelength, length)]
        super().__init__(layers)


class WaveguideChannels(Geometry):
    def __init__(
        self,
        params: Params = EMpyGeometryParameters(),
        width: float = 0.5,
        thickness: float = 0.22,
        length: float = 1,
        num_modes: int = 1,
        gap: float = 0.1,
        num_channels: int = 2,
        exclude_indices: list = [],
    ) -> None:

        # Create n
        starting_center = -0.5 * (num_channels - 1) * (gap + width)
        n_output = np.ones(params.mesh) * params.cladding_index
        for out in range(num_channels):
            if out not in exclude_indices:
                center = starting_center + out * (gap + width)
                left_edge = center - 0.5 * width
                right_edge = center + 0.5 * width
                n_output = np.where(
                    (left_edge <= params.x) * (params.x <= right_edge),
                    params.core_index,
                    n_output,
                )

        # Create modesolver
        output_channel = params.get_solver_index(thickness, num_modes, n_output)

        # Create layers
        self.layers = [Layer(output_channel, num_modes, params.wavelength, length)]
        super().__init__(self.layers)


class BraggGrating(Geometry):
    def __init__(
        self,
        params_left: Params = EMpyGeometryParameters(),
        params_right: Params = EMpyGeometryParameters(),
        width_left: float = 0.4,
        thickness_left: float = 0.22,
        length_left: float = 1,
        width_right: float = 0.6,
        thickness_right: float = 0.22,
        length_right: float = 1,
        num_modes: int = 1,
    ) -> None:

        # Create waveguides
        waveguide_left = Waveguide(
            params_left, width_left, thickness_left, length_left, num_modes
        )
        waveguide_right = Waveguide(
            params_right, width_right, thickness_right, length_right, num_modes
        )

        # Create layers
        self.layers = [*waveguide_left, *waveguide_right]
        super().__init__(self.layers)


class DirectionalCoupler(Geometry):
    def __init__(
        self,
        params: Params = EMpyGeometryParameters(),
        width: float = 0.5,
        thickness: float = 0.22,
        length: float = 25,
        gap: float = 0.2,
        num_modes: int = 1,
    ) -> None:

        # Create input waveguide channel
        input = WaveguideChannels(
            params, width, thickness, length, num_modes, gap, 2, exclude_indices=[1]
        )

        # Create main directional coupler
        coupler = WaveguideChannels(
            params, width, thickness, length, num_modes, gap, 2, exclude_indices=[]
        )

        # Create layers
        self.layers = [*input, *coupler]
        super().__init__(self.layers)



import numpy as np
import os
import EMpy_gpu
import collections
from emepy.materials import *
from typing import Callable
from shapely.geometry import Polygon, Point


def polygon_to_n_2D(
    polygon: "Polygon", x: list, y: list, subpixel=True, core_index=3.4, cladding_index=1.4
) -> "np.ndarray":
    # Create grid
    xx, yy = np.meshgrid(x, y)
    n = np.zeros(xx.shape)[:-1, :-1].T

    # Apply subpixel
    xlower, xupper = (x[:-1], x[1:])
    zlower, zupper = (y[:-1], y[1:])
    for i, xp in enumerate(zip(xlower, xupper)):
        for j, zp in enumerate(zip(zlower, zupper)):

            # Upper and lower points
            xl, xu = xp
            zl, zu = zp
            total_area = (xu - xl) * (zu - zl)

            # Create polygon of the pixel
            pixel_poly = Polygon([(xl, zl), (xl, zu), (xu, zu), (xu, zl)])

            # Get overlapping area
            overlapping_area = 0
            if pixel_poly.intersects(polygon):
                overlapping_area = pixel_poly.intersection(polygon).area

            # Calculate effective index
            if subpixel:
                n[i, j] = (
                    overlapping_area / total_area * core_index + (1 - overlapping_area / total_area) * cladding_index
                )
            elif overlapping_area:
                n[i, j] = core_index
            else:
                n[i, j] = cladding_index

    return n


def vertices_to_n(vertices: list, x: list, y: list, subpixel=True, core_index=3.4, cladding_index=1.4) -> "np.ndarray":
    """
    Takes vertices of a polygon and maps it to a grid using or not using subpixel smoothing
    """

    polygon = Polygon(vertices)
    return polygon_to_n_2D(polygon, x, y, subpixel, core_index, cladding_index)


def rectangle_to_n(
    center: tuple, width: float, thickness: float, x: list, y: list, subpixel=True, core_index=3.4, cladding_index=1.4
) -> "np.ndarray":

    xc, yc = center
    vertices = [
        (xc - width / 2, yc - thickness / 2),
        (xc - width / 2, yc + thickness / 2),
        (xc + width / 2, yc + thickness / 2),
        (xc + width / 2, yc - thickness / 2),
    ]
    return vertices_to_n(vertices, x, y, subpixel, core_index, cladding_index)


def circle_to_n(
    center: tuple, radius: float, x: list, y: list, subpixel=True, core_index=3.4, cladding_index=1.4
) -> "np.ndarray":

    polygon = Point(*center).buffer(radius)
    return polygon_to_n_2D(polygon, x, y, subpixel, core_index, cladding_index)

def interp2d(x, y, xx, yy, f, sci=False):

    def interp2d_partial(x, xp, fp):
        j = np.searchsorted(xp, x)
        j = np.clip(j, 0, len(xp) - 2)
        alpha = (x - xp[j]) / (xp[j + 1] - xp[j])
        alpha = np.repeat(alpha[:, np.newaxis], fp.shape[1], axis=1)
        return np.clip((fp[j, :] * alpha + fp[j+1, :] * (1 - alpha)), np.min(fp), np.max(fp))

    def on_real(x, y, xx, yy, f):

        if sci:
            from scipy.interpolate import griddata
            xxn, yyn = np.meshgrid(np.real(xx), np.real(yy))
            points = np.array((xxn.flatten(), yyn.flatten())).T
            n = f.flatten()
            xx, yy = np.meshgrid(np.real(x), np.real(y))
            return griddata(points, n, (xx, yy))
        else:
            f1 = interp2d_partial(x, xx, f)
            f2 = interp2d_partial(y, yy, f1.T)
            return f2.T

    return on_real(x, y, xx, yy, np.real(f)) + 1j * on_real(x, y, xx, yy, np.imag(f)) if np.iscomplexobj(f) else on_real(x, y, xx, yy, f)

class get_epsfunc(object):
    """Callable class for getting epsilon on a grid"""

    def __init__(
        self,
        width: float,
        thickness: float,
        cladding_width: float,
        cladding_thickness: float,
        core_index: float,
        cladding_index: float,
        compute: bool = False,
        profile: "np.ndarray" = None,
        nx: int = None,
        ny: int = None,
    ):
        """Returns the epsfunc for given parameters for a rectangular waveguide

        Parameters
        ----------
        width: float
            the width of the geometry
        thickness: float
            the thickess of the geometry
        cladding_width: float
            the width of the surrounding cladding (note this is total width and should be > core width)
        cladding_thickness: float
            the thickness of the surrounding cladding (note this is total thickness and should be > core thickness)
        core_index: float
            refractive index of the core
        cladding_index: float
            refractive index of the cladding
        compute: bool = False
            if true, will not place rectangle at center. This should only be necessary for compute_other_fields
        profile: "np.ndarray" = None
            the refractive index profile (note if providing a width, this should be left None)
        nx: int = None
            number of points in the x direction
        ny: int = None
            number of points in the y direction
        subpixel: bool = True
            if true, will use subpixel smoothing, assuming asking for a waveguide cross section and not providing an index map (recommended)

        Returns
        -------
        function
            an epsfunc function that takes an x,y grid arrays and returns the refractive index profile
        """
        self.width = width
        self.thickness = thickness
        self.cladding = cladding_width
        self.cladding = cladding_thickness
        self.core_index = core_index
        self.cladding_index = cladding_index
        self.compute = compute
        self.profile = profile
        self.nx = nx
        self.ny = ny

        if (width is not None) and (thickness is not None):
            self.epsfunction = self.epsfunc_2D_1

        elif (width is None) and (thickness is not None) and (profile is not None):
            self.epsfunction = self.epsfunc_2D_2

        elif (width is None) and (thickness is None) and (profile is not None):
            self.epsfunction = self.epsfunc_2D_3

        elif (thickness is None) and (width is not None):
            self.epsfunction = self.epsfunc_1D_1

        elif (thickness is None) and (width is None) and (profile is not None):
            self.epsfunction = self.epsfunc_1D_2

        else:
            raise Exception(
                "Need to provide width & thickness, or 1D profile and thickness, or 2D profile, or width for 1D, or 1D profile for 1D"
            )

    def __call__(self, x_, y_):
        return self.epsfunction(x_, y_)

    # Case 1 : width and thickness are defined
    def epsfunc_2D_1(self, x_, y_):
        """Return a matrix describing a 2d material.

        Parameters
        ----------
        x_: numpy array
            x values
        y_: numpy array
            y values

        Returns
        -------
        numpy array
            2d-matrix
        """
        xx, yy = np.meshgrid(x_, y_)
        if self.compute:
            n = np.where(
                (np.abs(np.real(xx.T) - self.cladding_width * 0.5) <= self.width * 0.5)
                * (np.abs(np.real(yy.T) - self.cladding_thickness * 0.5) <= self.thickness * 0.5),
                self.core_index ** 2 + 0j,
                self.cladding_index ** 2 + 0j,
            )
        else:
            n = np.where(
                (np.abs(np.real(xx.T)) <= self.width * 0.5) * (np.abs(np.real(yy.T)) <= self.thickness * 0.5),
                self.core_index ** 2 + 0j,
                self.cladding_index ** 2 + 0j,
            )
        return n

    # Case 2 : thickness and 1D n is defined
    def epsfunc_2D_2(self, x_, y_):

        n = self.profile
        xx, yy = np.meshgrid(np.real(x_), np.real(y_))
        n = np.interp(np.real(x_), np.real(self.nx), n).astype(complex)
        n = np.repeat(n, len(y_)).reshape((len(n), len(y_)))
        n = np.where((np.abs(np.real(yy.T)) <= self.thickness * 0.5), n, self.cladding_index + 0j) ** 2
        return n

    # Case 3 : 2D n is defined
    def epsfunc_2D_3(self, x_, y_):

        # from scipy.interpolate import griddata
        # xxn, yyn = np.meshgrid(np.real(self.nx), np.real(self.ny))
        # points = np.array((xxn.flatten(), yyn.flatten())).T
        # n = self.profile.flatten()
        # xx, yy = np.meshgrid(np.real(x_), np.real(y_))
        # n_real = griddata(points, np.real(n), (xx, yy))
        # n_imag = griddata(points, np.imag(n), (xx, yy))
        # n = (n_real + 1j * n_imag) ** 2

        return interp2d(x_, y_, self.nx, self.ny, self.profile) ** 2

    # Case 4: width only
    def epsfunc_1D_1(self, x_, y_):

        n = np.where(
            (np.abs(np.real(x_)) <= self.width * 0.5), self.core_index ** 2 + 0j, self.cladding_index ** 2 + 0j
        ).reshape(len(x_), 1)

        return n

    # Case 5: 1D n only
    def epsfunc_1D_2(self, x_, y_):

        n = np.interp(np.real(x_), np.real(self.nx), self.profile).astype(complex).reshape(len(x_), 1)
        return n


def get_isotropic_epsfunc(
    epsfunc_xx, epsfunc_yy, epsfunc_zz, epsfunc_xy=None, epsfunc_yx=None
) -> Callable[["np.ndarray", "np.ndarray"], "np.ndarray"]:
    """Returns an epsfunction for an isotropic medium"""

    def epsfunc_iso(x_, y_):
        xx = epsfunc_xx(x_, y_)
        yy = epsfunc_yy(x_, y_)
        zz = epsfunc_zz(x_, y_)
        xy = epsfunc_xy(x_, y_) if epsfunc_xy is not None else 0 * xx
        yx = epsfunc_yx(x_, y_) if epsfunc_yx is not None else 0 * xx

        return np.array([xx, xy, yx, yy, zz])

    return epsfunc_iso


def create_polygon(x: "np.ndarray", y: "np.ndarray", n: "np.ndarray", detranslate: bool = True) -> list:
    """Given a grid and a refractive index profile, will return the vertices of the polygon for importing into libraries such as Lumerical

    Parameters
    ----------
    x : "np.ndarray"
        the x grid
    y : "np.ndarray"
        the y grid
    n :"np.ndarray"
        the refractive index profile
    detranslate : bool
        if True, will detranslate the vertices

    Returns
    -------
    list[tuples]
        the resulting vertices

    """

    x0, y0 = [x.copy(), y.copy()]
    diff = np.abs(np.diff(n, axis=1))
    where = np.argwhere(diff > np.mean(n))
    tx = np.mean(x[where[:, 0]])
    ty = np.mean(y[where[:, 1]])
    x0 -= tx
    y0 -= ty
    diff = diff[1:, :] + diff[:-1, :]
    diff2 = np.abs(np.diff(n, axis=0))
    diff2 = diff2[:, 1:] + diff2[:, :-1]
    diff = diff + diff2
    diff = np.where(diff, 1, 0)

    newd = {}
    for i in np.argwhere(diff):
        angle = np.angle(x0[i[0]] + 1j * y0[i[1]])
        if detranslate:
            newd[angle] = [y0[i[1]] - ty, x0[i[0]] - tx]
        else:
            newd[angle] = [y0[i[1]], x0[i[0]]]

    od = collections.OrderedDict(sorted(newd.items()))

    if detranslate:
        return np.array(list(od.values())).astype(float)
    else:
        return [np.array(list(od.values())).astype(float), tx, ty]


def interp(
    x: "np.ndarray", y: "np.ndarray", x0: "np.ndarray", y0: "np.ndarray", f: "np.ndarray", centered: bool
) -> "np.ndarray":
    """Interpolate a 2D complex array.

    Parameters
    ----------
    x:"np.ndarray"
        the new x grid
    y:"np.ndarray"
        the new y grid
    x0:"np.ndarray"
        the original x grid
    y0:"np.ndarray"
        the original y grid
    f:"np.ndarray"
        the field to interpolate
    centered:bool
        whether or not it needs to stil be shifted

    Returns
    -------
    np.ndarray
        the interpolated field
    """

    if centered:
        # electric fields and intensity are centered
        x0 = EMpy_gpu.utils.centered1d(x0)
        y0 = EMpy_gpu.utils.centered1d(y0)

    f1r = np.zeros((len(x0), len(y)))
    f1i = np.zeros((len(x0), len(y)))
    for ix0 in range(len(x0)):
        f1r[ix0, :] = np.interp(y, y0, np.real(f[ix0, :]))
        f1i[ix0, :] = np.interp(y, y0, np.imag(f[ix0, :]))
    fr = np.zeros((len(x), len(y)))
    fi = np.zeros((len(x), len(y)))
    for iy in range(len(y)):
        fr[:, iy] = np.interp(x, x0, f1r[:, iy])
        fi[:, iy] = np.interp(x, x0, f1i[:, iy])
    return fr + 1j * fi


def interp1d(x: "np.ndarray", x0: "np.ndarray", f: "np.ndarray", centered: bool) -> "np.ndarray":
    """Interpolate a 1D complex array.

    Parameters
    ----------
    x:"np.ndarray"
        the new grid
    x0:"np.ndarray"
        the original grid
    f:"np.ndarray"
        the field to interpolate
    centered:bool
        whether or not it needs to stil be shifted

    Returns
    -------
    np.ndarray
        the interpolated field
    """

    if centered:
        # electric fields and intensity are centered
        x0 = EMpy_gpu.utils.centered1d(x0)

    f1r = np.interp(x, x0, np.real(f[:]))
    f1i = np.interp(x, x0, np.imag(f[:]))
    return f1r + 1j * f1i


def into_chunks(location: str, name: str, chunk_size: int = 20000000) -> None:
    """Takes a large serialized file and breaks it up into smaller chunk files

    Parameters
    ---------
    location : string
        the absolute or relative path of the large file
    name : string
        the name of the serialized smaller components (will have _chunk_# appended to it)
    chunk_size : int
        how big each save chunk should be
    """
    CHUNK_SIZE = chunk_size
    f = open(location, "rb")
    chunk = f.read(CHUNK_SIZE)
    count = 0
    while chunk:  # loop until the chunk is empty (the file is exhausted)
        with open(name + "_chunk_" + str(count), "wb+") as w:
            w.write(chunk)
        count += 1
        chunk = f.read(CHUNK_SIZE)  # read the next chunk
    f.close()


def from_chunks(location: str, name: str) -> None:
    """Takes a directory of serialized chunks that were made using into_chunks and combines them back into a large serialized file

    Parameters
    ----------
    location : string
        the path of the directory where the chunks are located
    name : string
        the name of the serialized file to create (make sure to include file extension if it matters)
    """
    if location[-1] != "/":
        location += "/"
    f = open(name, "wb+")
    direc = os.listdir(location)
    keys = [int(d[9:]) for d in direc]
    dic = dict(zip(keys, direc))
    for i in sorted(dic):
        f.write(open(location + dic[i], "rb").read())
    f.close()


def _get_eps(
    xc: "np.ndarray", yc: "np.ndarray", epsfunc: Callable[["np.ndarray", "np.ndarray"], "np.ndarray"]
) -> tuple:
    """Used by compute_other_fields and adapted from the EMpy library"""
    tmp = epsfunc(xc, yc)

    def _reshape(tmp):
        """pads the array by duplicating edge values"""
        tmp = np.c_[tmp[:, 0:1], tmp, tmp[:, -1:]]
        tmp = np.r_[tmp[0:1, :], tmp, tmp[-1:, :]]
        return tmp

    if tmp.ndim == 2:
        tmp = _reshape(tmp)
        epsxx = epsyy = epszz = tmp
        epsxy = epsyx = np.zeros_like(epsxx)

    elif tmp.ndim == 3:
        assert tmp.shape[2] == 5, "eps must be NxMx5"
        epsxx = _reshape(tmp[:, :, 0])
        epsxy = _reshape(tmp[:, :, 1])
        epsyx = _reshape(tmp[:, :, 2])
        epsyy = _reshape(tmp[:, :, 3])
        epszz = _reshape(tmp[:, :, 4])

    else:
        raise ValueError("Invalid eps")

    return epsxx, epsxy, epsyx, epsyy, epszz




import io
import os
import re

from setuptools import find_packages
from setuptools import setup

__version__ = "1.2.2"

def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type(u"")
    with io.open(filename, mode="r", encoding="utf-8") as fd:
        return re.sub(text_type(r":[a-z]+:`~?(.*?)`"), text_type(r"``\1``"), fd.read())


direct = []
# "electromagneticpython @ git+git://github.com/lbolla/EMpy@4bf1b01#egg=electromagneticpython",
# "electromagneticpythongpu @ git+git://github.com/hammy4815/EMpy_gpu@master#egg=electromagneticpythongpu",


def get_install_requires():
    with open("requirements.txt", "r") as f:
        return direct + [line.strip() for line in f.readlines() if not line.startswith("-")]

long_description = "EMEPy - Eigenmode Expansion Python. An open-source Python package for the simulation of electromagnetic devices."

setup(
    name="emepy",
    version=__version__,
    url="https://github.com/BYUCamachoLab/emepy",
    license="MIT",
    author="Ian Hammond",
    author_email="ihammond@byu.edu",
    description="Eigenmode Expansion Python",
    long_description=long_description,
    packages=find_packages(exclude=("tests",)),
    install_requires=get_install_requires(),
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
    ],
)
